#! /usr/bin/env python

"""
Module with functions for computing SVDs.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['svd_wrapper',
           'randomized_svd_gpu',
           'get_eigenvectors']

import warnings
try:
    import cupy
    no_cupy = False
except ImportError:
    msg = "Cupy not found. Have a GPU? Consider setting up a CUDA environment "
    msg += "and installing cupy >= 2.0.0"
    warnings.warn(msg, ImportWarning)
    no_cupy = True
try:
    import torch
    no_torch = False
except ImportError:
    msg = "Pytorch not found. Have a GPU? Consider setting up a CUDA "
    msg += "environment and installing pytorch"
    warnings.warn(msg, ImportWarning)
    no_torch = True

import numpy as np
from numpy import linalg
from matplotlib import pyplot as plt 
from scipy.sparse.linalg import svds
from sklearn.decomposition import randomized_svd
from sklearn.metrics import mean_squared_error as MSE
from sklearn.metrics import mean_absolute_error as MAE
from sklearn.utils import check_random_state
from ..var import matrix_scaling


def svd_wrapper(matrix, mode, ncomp, debug, verbose, usv=False,
                random_state=None, to_numpy=True):
    """ Wrapper for different SVD libraries (CPU and GPU). 
      
    Parameters
    ----------
    matrix : array_like, 2d
        2d input matrix.
    mode : {'lapack', 'arpack', 'eigen', 'randsvd', 'cupy', 'eigencupy',
            'randcupy', 'pytorch', 'eigenpytorch', 'randpytorch'}, str optional
        Switch for the SVD method/library to be used. ``lapack`` uses the LAPACK 
        linear algebra library through Numpy and it is the most conventional way 
        of computing the SVD (deterministic result computed on CPU). ``arpack`` 
        uses the ARPACK Fortran libraries accessible through Scipy (computation
        on CPU). ``eigen`` computes the singular vectors through the 
        eigendecomposition of the covariance M.M' (computation on CPU).
        ``randsvd`` uses the randomized_svd algorithm implemented in Sklearn 
        (computation on CPU). ``cupy`` uses the Cupy library for GPU computation
        of the SVD as in the LAPACK version. ``eigencupy`` offers the same 
        method as with the ``eigen`` option but on GPU (through Cupy). 
        ``randcupy`` is an adaptation of the randomized_svd algorithm, where all
        the computations are done on a GPU (through Cupy). ``pytorch`` uses the
        Pytorch library for GPU computation of the SVD. ``eigenpytorch`` offers
        the same method as with the ``eigen`` option but on GPU (through
        Pytorch). ``randpytorch`` is an adaptation of the randomized_svd
        algorithm, where all the linear algebra computations are done on a GPU
        (through Pytorch).
    ncomp : int
        Number of singular vectors to be obtained. In the cases when the full
        SVD is computed (LAPACK, ARPACK, EIGEN, CUPY), the matrix of singular 
        vectors is truncated. 
    debug : bool
        If True the explained variance ratio is computed and displayed.
    verbose: bool
        If True intermediate information is printed out.
    usv : bool optional
        If True the 3 terms of the SVD factorization are returned.
    random_state : int, RandomState instance or None, optional
        If int, random_state is the seed used by the random number generator.
        If RandomState instance, random_state is the random number generator.
        If None, the random number generator is the RandomState instance used
        by np.random. Used for ``randsvd`` mode.
    to_numpy : bool, optional
        If True (by default) the arrays computed in GPU are transferred from
        VRAM and converted to numpy ndarrays.

    Returns
    -------
    The right singular vectors of the input matrix. If ``usv`` is True it 
    returns the left and right singular vectors and the singular values of the
    input matrix.
    
    References
    ----------
    * For ``lapack`` SVD mode see:
    https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.linalg.svd.html
    http://www.netlib.org/lapack/
    
    * For ``eigen`` mode see:
    https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.linalg.eigh.html
    
    * For ``arpack`` SVD mode see:
    https://docs.scipy.org/doc/scipy-0.19.1/reference/generated/scipy.sparse.linalg.svds.html
    http://www.caam.rice.edu/software/ARPACK/
    
    * For ``randsvd`` SVD mode see:
    https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/extmath.py
    Finding structure with randomness: Stochastic algorithms for constructing
    approximate matrix decompositions
    Halko, et al., 2009 http://arxiv.org/abs/arXiv:0909.4061
    
    * For ``cupy`` SVD mode see:
    https://docs-cupy.chainer.org/en/stable/reference/generated/cupy.linalg.svd.html
    
    * For ``eigencupy`` mode see:
    https://docs-cupy.chainer.org/en/master/reference/generated/cupy.linalg.eigh.html

    * For ``pytorch`` SVD mode see:
    http://pytorch.org/docs/master/torch.html#torch.svd

    * For ``eigenpytorch`` mode see:
    http://pytorch.org/docs/master/torch.html#torch.eig

    """

    def reconstruction(ncomp, U, S, V, var=1):
        if mode == 'lapack':
            rec_matrix = np.dot(U[:, :ncomp],
                                np.dot(np.diag(S[:ncomp]), V[:ncomp]))
            rec_matrix = rec_matrix.T
            print('  Matrix reconstruction with {:} PCs:'.format(ncomp))
            print('  Mean Absolute Error =', MAE(matrix, rec_matrix))
            print('  Mean Squared Error =', MSE(matrix, rec_matrix))

            # see https://github.com/scikit-learn/scikit-learn/blob/c3980bcbabd9d2527548820581725df2904e4a0d/sklearn/decomposition/pca.py
            exp_var = (S ** 2) / (S.shape[0] - 1)
            full_var = np.sum(exp_var)
            explained_variance_ratio = exp_var / full_var   # % of variance explained by each PC
            ratio_cumsum = np.cumsum(explained_variance_ratio)
        elif mode == 'eigen':
            exp_var = (S ** 2) / (S.shape[0] - 1)
            full_var = np.sum(exp_var)
            explained_variance_ratio = exp_var / full_var   # % of variance explained by each PC
            ratio_cumsum = np.cumsum(explained_variance_ratio)
        else:
            rec_matrix = np.dot(U, np.dot(np.diag(S), V))
            print('  Matrix reconstruction MAE =', MAE(matrix, rec_matrix))
            exp_var = (S ** 2) / (S.shape[0] - 1)
            full_var = np.var(matrix, axis=0).sum()
            explained_variance_ratio = exp_var / full_var   # % of variance explained by each PC
            if var == 1:
                pass
            else:
                explained_variance_ratio = explained_variance_ratio[::-1]
            ratio_cumsum = np.cumsum(explained_variance_ratio)
            msg = '  This info makes sense when the matrix is mean centered '
            msg += '(temp-mean scaling)'
            print(msg)

        lw = 2; alpha = 0.4
        fig = plt.figure(figsize=(6, 3))
        fig.subplots_adjust(wspace=0.4)
        ax1 = plt.subplot2grid((1, 3), (0, 0), colspan=2)
        ax1.step(range(explained_variance_ratio.shape[0]),
                 explained_variance_ratio, alpha=alpha, where='mid',
                 label='Individual EVR', lw=lw)
        ax1.plot(ratio_cumsum, '.-', alpha=alpha,
                 label='Cumulative EVR', lw=lw)
        ax1.legend(loc='best', frameon=False, fontsize='medium')
        ax1.set_ylabel('Explained variance ratio (EVR)')
        ax1.set_xlabel('Principal components')
        ax1.grid(linestyle='solid', alpha=0.2)
        ax1.set_xlim(-10, explained_variance_ratio.shape[0] + 10)
        ax1.set_ylim(0, 1)

        trunc = 20
        ax2 = plt.subplot2grid((1, 3), (0, 2), colspan=1)
        # plt.setp(ax2.get_yticklabels(), visible=False)
        ax2.step(range(trunc), explained_variance_ratio[:trunc], alpha=alpha,
                 where='mid', lw=lw)
        ax2.plot(ratio_cumsum[:trunc], '.-', alpha=alpha, lw=lw)
        ax2.set_xlabel('Principal components')
        ax2.grid(linestyle='solid', alpha=0.2)
        ax2.set_xlim(-2, trunc + 2)
        ax2.set_ylim(0, 1)

        msg = '  Cumulative explained variance ratio for {:} PCs = {:.5f}'
        # plt.savefig('figure.pdf', dpi=300, bbox_inches='tight')
        print(msg.format(ncomp, ratio_cumsum[ncomp - 1]))

    # --------------------------------------------------------------------------

    if not matrix.ndim == 2:
        raise TypeError('Input matrix is not a 2d array')

    if usv:
        if mode not in ('lapack', 'arpack', 'randsvd', 'cupy', 'randcupy',
                        'pytorch', 'randpytorch'):
            msg = "Returning USV is supported with modes lapack, arpack, "
            msg += "randsvd, cupy, randcupy, pytorch or randpytorch"
            raise ValueError(msg)

    if ncomp > min(matrix.shape[0], matrix.shape[1]):
        msg = '{:} PCs cannot be obtained from a matrix with size [{:},{:}].'
        msg += ' Increase the size of the patches or request less PCs'
        raise RuntimeError(msg.format(ncomp, matrix.shape[0], matrix.shape[1]))

    if mode == 'eigen':
        # building the covariance as np.dot(matrix.T,matrix) is slower and takes more memory
        C = np.dot(matrix, matrix.T)        # covariance matrix
        e, EV = linalg.eigh(C)              # eigenvalues and eigenvectors
        pc = np.dot(EV.T, matrix)           # PCs using a compact trick when cov is MM'
        V = pc[::-1]                        # reverse since last eigenvectors are the ones we want
        S = np.sqrt(e)[::-1]                # reverse since eigenvalues are in increasing order
        if debug:
            reconstruction(ncomp, None, S, None)
        for i in range(V.shape[1]):
            V[:, i] /= S                    # scaling by the square root of eigenvalues
        V = V[:ncomp]
        if verbose:
            print('Done PCA with numpy linalg eigh functions')

    elif mode == 'lapack':
        # n_frames is usually smaller than n_pixels. In this setting taking the SVD of M'
        # and keeping the left (transposed) SVs is faster than taking the SVD of M (right SVs)
        U, S, V = linalg.svd(matrix.T, full_matrices=False)
        if debug:
            reconstruction(ncomp, U, S, V)
        V = V[:ncomp]                       # we cut projection matrix according to the # of PCs
        U = U[:, :ncomp]
        S = S[:ncomp]
        if verbose:
            print('Done SVD/PCA with numpy SVD (LAPACK)')

    elif mode == 'arpack':
        U, S, V = svds(matrix, k=ncomp)
        if debug:
            reconstruction(ncomp, U, S, V, -1)
        if verbose:
            print('Done SVD/PCA with scipy sparse SVD (ARPACK)')

    elif mode == 'randsvd':
        U, S, V = randomized_svd(matrix, n_components=ncomp, n_iter=2,
                                 transpose='auto', random_state=random_state)
        if debug:
            reconstruction(ncomp, U, S, V)
        if verbose:
            print('Done SVD/PCA with randomized SVD')

    elif mode == 'cupy':
        if no_cupy:
            raise RuntimeError('Cupy is not installed')
        a_gpu = cupy.array(matrix)
        a_gpu = cupy.asarray(a_gpu)  # move the data to the current device
        u_gpu, s_gpu, vh_gpu = cupy.linalg.svd(a_gpu, full_matrices=True,
                                               compute_uv=True)
        V = vh_gpu[:ncomp]
        if to_numpy:
            V = cupy.asnumpy(V)
        if usv:
            S = s_gpu[:ncomp]
            if to_numpy:
                S = cupy.asnumpy(S)
            U = u_gpu[:, :ncomp]
            if to_numpy:
                U = cupy.asnumpy(U)
        if verbose:
            print('Done SVD/PCA with cupy (GPU)')

    elif mode == 'randcupy':
        if no_cupy:
            raise RuntimeError('Cupy is not installed')
        U, S, V = randomized_svd_gpu(matrix, ncomp, n_iter=2, lib='cupy')
        if to_numpy:
            V = cupy.asnumpy(V)
            S = cupy.asnumpy(S)
            U = cupy.asnumpy(U)
        if debug:
            reconstruction(ncomp, U, S, V)
        if verbose:
            print('Done randomized SVD/PCA with cupy (GPU)')

    elif mode == 'eigencupy':
        if no_cupy:
            raise RuntimeError('Cupy is not installed')
        a_gpu = cupy.array(matrix)
        a_gpu = cupy.asarray(a_gpu)         # move the data to the current device
        C = cupy.dot(a_gpu, a_gpu.T)        # covariance matrix
        e, EV = cupy.linalg.eigh(C)         # eigenvalues and eigenvectors
        pc = cupy.dot(EV.T, a_gpu)          # PCs using a compact trick when cov is MM'
        V = pc[::-1]                        # reverse since last eigenvectors are the ones we want
        S = cupy.sqrt(e)[::-1]              # reverse since eigenvalues are in increasing order
        if debug:
            reconstruction(ncomp, None, S, None)
        for i in range(V.shape[1]):
            V[:, i] /= S                    # scaling by the square root of eigenvalues
        V = V[:ncomp]
        if to_numpy:
            V = cupy.asnumpy(V)
        if verbose:
            print('Done PCA with cupy eigh function (GPU)')

    elif mode == 'pytorch':
        if no_torch:
            raise RuntimeError('Pytorch is not installed')
        a_gpu = torch.Tensor.cuda(torch.from_numpy(matrix.astype('float32').T))
        u_gpu, s_gpu, vh_gpu = torch.svd(a_gpu)
        V = vh_gpu[:ncomp]
        S = s_gpu[:ncomp]
        U = torch.transpose(u_gpu, 0, 1)[:ncomp]
        if to_numpy:
            V = np.array(V)
            S = np.array(S)
            U = np.array(U)
        if verbose:
            print('Done SVD/PCA with pytorch (GPU)')

    elif mode == 'eigenpytorch':
        if no_torch:
            raise RuntimeError('Pytorch is not installed')
        a_gpu = torch.Tensor.cuda(torch.from_numpy(matrix.astype('float32')))
        C = torch.mm(a_gpu, torch.transpose(a_gpu, 0, 1))
        e, EV = torch.eig(C, eigenvectors=True)
        V = torch.mm(torch.transpose(EV, 0, 1), a_gpu)
        S = torch.sqrt(e[:, 0])
        if debug:
            reconstruction(ncomp, None, S, None)
        for i in range(V.shape[1]):
            V[:, i] /= S
        V = V[:ncomp]
        if to_numpy:
            V = np.array(V)
        if verbose:
            print('Done PCA with pytorch eig function')

    elif mode == 'randpytorch':
        if no_torch:
            raise RuntimeError('Pytorch is not installed')
        U, S, V = randomized_svd_gpu(matrix, ncomp, n_iter=2, lib='pytorch')
        if to_numpy:
            V = np.array(V)
            S = np.array(S)
            U = np.array(U)
        if debug:
            reconstruction(ncomp, U, S, V)
        if verbose:
            print('Done randomized SVD/PCA with randomized pytorch (GPU)')

    else:
        raise ValueError('The SVD mode is not available')

    if usv:
        if mode == 'lapack':
            return V.T, S, U.T
        elif mode == 'pytorch':
            if to_numpy:
                return V.T, S, U.T
            else:
                return torch.transpose(V, 0, 1), S, torch.transpose(U, 0, 1)
        else:
            return U, S, V
    else:
        if mode == 'lapack':
            return U.T
        elif mode == 'pytorch':
            return U
        else:
            return V


def get_eigenvectors(ncomp, data, svd_mode, mode='noise', noise_error=1e-3,
                     cevr=0.9, max_evs=None, data_ref=None, debug=False):
    """ Getting ``ncomp`` eigenvectors. Choosing the size of the PCA truncation
    when ``ncomp`` is set to None.
    """
    no_dataref = False
    if data_ref is None:
        no_dataref = True
        data_ref = data

    if max_evs is None:
        max_evs = min(data_ref.shape[0], data_ref.shape[1])

    if ncomp is None:
        ncomp = 0
        V_big = svd_wrapper(data_ref, svd_mode, max_evs, False, False)

        if mode=='noise':
            data_ref_sc = matrix_scaling(data_ref, 'temp-standard')
            data_sc = matrix_scaling(data, 'temp-standard')
            V_sc = svd_wrapper(data_ref_sc, svd_mode, max_evs, False, False)

            px_noise = []
            px_noise_decay = 1
            # Noise (px stddev of residuals) to be lower than a given threshold
            while px_noise_decay >= noise_error:
                ncomp += 1
                V = V_sc[:ncomp]
                if no_dataref:
                    transformed = np.dot(data_sc, V.T)
                    reconstructed = np.dot(transformed, V)
                else:
                    transformed = np.dot(V, data_sc)
                    reconstructed = np.dot(transformed.T, V).T
                residuals = data_sc - reconstructed
                curr_noise = np.std((np.median(residuals, axis=0)))
                px_noise.append(curr_noise)
                if ncomp > 1:
                    px_noise_decay = px_noise[-2] - curr_noise
                # print '{:} {:.4f} {:.4f}'.format(ncomp, curr_noise, px_noise_decay)
            V = V_big[:ncomp]

        elif mode=='cevr':
            data_sc = matrix_scaling(data, 'temp-mean')
            _, S, _ = svd_wrapper(data_sc, svd_mode, min(data_sc.shape[0],
                                                         data_sc.shape[1]),
                                  False, False, usv=True)
            exp_var = (S ** 2) / (S.shape[0] - 1)
            full_var = np.sum(exp_var)
            explained_variance_ratio = exp_var / full_var  # % of variance explained by each PC
            ratio_cumsum = np.cumsum(explained_variance_ratio)
            ncomp = np.searchsorted(ratio_cumsum, cevr) + 1
            V = V_big[:ncomp]

        if debug:
            print('ncomp', ncomp)

    else:
        # Performing SVD/PCA according to "svd_mode" flag
        V = svd_wrapper(data_ref, svd_mode, ncomp, debug=False, verbose=False)

    return V


def randomized_svd_gpu(M, n_components, n_oversamples=10, n_iter='auto',
                       transpose='auto', random_state=0, lib='cupy'):
    """Computes a truncated randomized SVD on GPU. Adapted from Sklearn.

    Parameters
    ----------
    M : ndarray or sparse matrix
        Matrix to decompose
    n_components : int
        Number of singular values and vectors to extract.
    n_oversamples : int (default is 10)
        Additional number of random vectors to sample the range of M so as
        to ensure proper conditioning. The total number of random vectors
        used to find the range of M is n_components + n_oversamples. Smaller
        number can improve speed but can negatively impact the quality of
        approximation of singular vectors and singular values.
    n_iter : int or 'auto' (default is 'auto')
        Number of power iterations. It can be used to deal with very noisy
        problems. When 'auto', it is set to 4, unless `n_components` is small
        (< .1 * min(X.shape)) `n_iter` in which case is set to 7.
        This improves precision with few components.
    transpose : True, False or 'auto' (default)
        Whether the algorithm should be applied to M.T instead of M. The
        result should approximately be the same. The 'auto' mode will
        trigger the transposition if M.shape[1] > M.shape[0] since this
        implementation of randomized SVD tend to be a little faster in that
        case.
    random_state : int, RandomState instance or None, optional (default=None)
        The seed of the pseudo random number generator to use when shuffling
        the data.  If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the random number
        generator; If None, the random number generator is the RandomState
        instance used by `np.random`.
    lib : {'cupy', 'pytorch'}, str optional
        Chooses the GPU library to be used.

    Notes
    -----
    This algorithm finds a (usually very good) approximate truncated
    singular value decomposition using randomization to speed up the
    computations. It is particularly fast on large matrices on which
    you wish to extract only a small number of components. In order to
    obtain further speed up, `n_iter` can be set <=2 (at the cost of
    loss of precision).

    References
    ----------
    * Finding structure with randomness: Stochastic algorithms for constructing
      approximate matrix decompositions
      Halko, et al., 2009 http://arxiv.org/abs/arXiv:0909.4061
    * A randomized algorithm for the decomposition of matrices
      Per-Gunnar Martinsson, Vladimir Rokhlin and Mark Tygert
    * An implementation of a randomized algorithm for principal component
      analysis
      A. Szlam et al. 2014
    """
    random_state = check_random_state(random_state)
    n_random = n_components + n_oversamples
    n_samples, n_features = M.shape

    if n_iter == 'auto':
        # Checks if the number of iterations is explicitly specified
        n_iter = 7 if n_components < .1 * min(M.shape) else 4

    if transpose == 'auto':
        transpose = n_samples < n_features
    if transpose:
        M = M.T # this implementation is a bit faster with smaller shape[1]

    if lib == 'cupy':
        M = cupy.array(M)
        M = cupy.asarray(M)

        # Generating normal random vectors with shape: (M.shape[1], n_random)
        Q = random_state.normal(size=(M.shape[1], n_random))
        Q = cupy.array(Q)
        Q = cupy.asarray(Q)

        # Perform power iterations with Q to further 'imprint' the top
        # singular vectors of M in Q
        for i in range(n_iter):
            Q = cupy.dot(M, Q)
            Q = cupy.dot(M.T, Q)

        # Sample the range of M using by linear projection of Q. Extract an orthonormal basis
        Q, _ = cupy.linalg.qr(cupy.dot(M, Q), mode='reduced')

        # project M to the (k + p) dimensional space using the basis vectors
        B = cupy.dot(Q.T, M)

        B = cupy.array(B)
        Q = cupy.array(Q)
        # compute the SVD on the thin matrix: (k + p) wide
        Uhat, s, V = cupy.linalg.svd(B, full_matrices=False, compute_uv=True)
        del B
        U = cupy.dot(Q, Uhat)

        if transpose:
            # transpose back the results according to the input convention
            return V[:n_components, :].T, s[:n_components], U[:,
                                                            :n_components].T
        else:
            return U[:, :n_components], s[:n_components], V[:n_components, :]

    elif lib == 'pytorch':
        M_gpu = torch.Tensor.cuda(torch.from_numpy(M.astype('float32')))

        # Generating normal random vectors with shape: (M.shape[1], n_random)
        Q = torch.cuda.FloatTensor(M_gpu.shape[1], n_random).normal_()

        # Perform power iterations with Q to further 'imprint' the top
        # singular vectors of M in Q
        for i in range(n_iter):
            Q = torch.mm(M_gpu, Q)
            Q = torch.mm(torch.transpose(M_gpu, 0, 1), Q)

        # Sample the range of M using by linear projection of Q. Extract an orthonormal basis
        Q, _ = torch.qr(torch.mm(M_gpu, Q))

        # project M to the (k + p) dimensional space using the basis vectors
        B = torch.mm(torch.transpose(Q, 0, 1), M_gpu)

        # compute the SVD on the thin matrix: (k + p) wide
        Uhat, s, V = torch.svd(B)
        del B
        U = torch.mm(Q, Uhat)

        if transpose:
            # transpose back the results according to the input convention
            return (torch.transpose(V[:n_components, :], 0, 1),
                    s[:n_components],
                    torch.transpose(U[:, :n_components], 0, 1))
        else:
            return U[:, :n_components], s[:n_components], V[:n_components, :]




"""
Subpackage ``LLSG`` contains the first version of our Local Low-rank plus Sparse
plus Gaussian-noise decomposition (Gomez Gonzalez et al. 2016) for ADI data.
"""
from __future__ import absolute_import

from .llsg import *
from .thresholding import *


#! /usr/bin/env python

"""
Module with various functions.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez, O. Wertz'
__all__ = ['pp_subplots',
           'plot_surface',
           'get_fwhm',
           'lines_of_code']

import os
import numpy as np
from matplotlib.pyplot import (figure, subplot, show, colorbar, rc, axes,
                               Circle, savefig)
import matplotlib.colors as colors
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.axes_grid1 import make_axes_locatable
from .shapes import frame_center


def get_fwhm(lambd, diameter, pxscale):
    """ Returnes the instrument FWHM [px] given the wavelenght [m], diameter [m] 
    and plate/pixel scale [arcs/px]. In vip_hci/conf/param.py can be found
    dictionaries with the parameters for different instruments.                           
    """
    fwhm = lambd/diameter*206265/pxscale                    
    return fwhm 


def pp_subplots(*args, **kwargs):
    """ Wrapper for easy creation of pyplot subplots. It is convenient for 
    displaying VIP images in jupyter notebooks. 
    
    Parameters in **kwargs
    ----------------------
    angscale : axes in angular scale (arcsecs)
    angticksep : separation for the ticks when using axis in angular scale
    arrow : to show an arrow pointing to input px coordinates
    arrowalpha : alpha transparency for the arrow
    arrowlength : length of the arrow, 20 px by default
    arrowshiftx : shift in x of the arrow pointing position, 5 px by default
    axis : show the axis, on by default
    circle : to show a circle at given px coordinates, list of tuples
    circlerad : radius of the circle, 6 px by default
    cmap : colormap to be used, 'viridis' by default
    colorb : to attach a colorbar, on by default
    cross : to show a crosshair at given px coordinates
    crossalpha : alpha transparency of thr crosshair
    dpi : dots per inch, for plot quality
    getfig : returns the figure
    grid : for showing a grid over the image, off by default
    gridalpha : alpha transparency of the grid
    gridcolor : color of the grid lines
    gridspacing : separation of the grid lines in pixels
    horsp : horizontal gap between subplots
    label : text for annotating on subplots
    labelpad : padding of the label from the left bottom corner
    labelsize : size of the labels
    log : log colorscale
    maxplot : sets the maximum number of subplots when the input is a 3d array
    pxscale : pixel scale in arcseconds/px. Default 0.01 for Keck/NIRC2
    rows : how many rows (subplots in a grid)
    save : if a string is provided the plot is saved using this as the path
    showcent : to show a big crosshair at the center of the frame 
    title : title of the plot(s), None by default
    vmax : for stretching the displayed pixels values
    vmin : for stretching the displayed pixels values
    versp : vertical gap between subplots

    """
    parlist = ['angscale', 'angticksep', 'arrow', 'arrowalpha', 'arrowlength',
               'arrowshiftx', 'axis', 'circle', 'circlealpha', 'circlerad',
               'cmap', 'colorb', 'cross', 'crossalpha', 'dpi', 'getfig', 'grid',
               'gridalpha', 'gridcolor', 'gridspacing', 'horsp', 'label',
               'labelpad', 'labelsize', 'log', 'maxplots', 'pxscale', 'rows',
               'save', 'showcent', 'title', 'vmax', 'vmin', 'versp']
    
    for key in kwargs.keys():
        if key not in parlist:
            print("Parameter '{:}' not recognized".format(key))
            print("Available parameters are: {:}".format(parlist))

    # GEOM ---------------------------------------------------------------------
    num_plots = len(args)
    if num_plots == 1:
        if args[0].ndim == 3:
            data = args[0]
            if 'maxplots' in kwargs:
                maxplots = kwargs['maxplots']
            else:
                maxplots = 10
            num_plots = min(data.shape[0], maxplots)
        else:
            data = args
    elif num_plots > 1:
        data = args
        for i in range(num_plots):
            if not args[i].ndim == 2:
                msg = "Wrong input. Must be either several 2d arrays (images) "
                msg += "or a single 3d array"
                raise TypeError(msg)

    if 'rows' in kwargs:
        if not isinstance(kwargs['rows'], int):
            raise TypeError
        else:
            rows = kwargs['rows']
    else:
        rows = 1

    if num_plots % rows == 0:
        cols = num_plots / rows
    else:
        cols = (num_plots / rows) + 1

    # CIRCLE -------------------------------------------------------------------
    if 'circle' in kwargs:
        if not isinstance(kwargs['circle'], list) and isinstance(kwargs['circle'], tuple):
            show_circle = True
            coor_circle = [kwargs['circle'] for i in range(num_plots)]
        else:
            if not isinstance(kwargs['circle'][0], tuple):
                print("Circle must be a tuple (X,Y) or list of tuples (X,Y)")
                show_circle = False
            else:
                show_circle = True
                coor_circle = kwargs['circle']
    else:
        show_circle = False

    if 'circlerad' in kwargs:
        circle_rad = kwargs['circlerad']
    else:
        circle_rad = 6

    if 'circlealpha' in kwargs:
        circle_alpha = kwargs['circlealpha']
    else:
        circle_alpha = 0.8

    # ARROW --------------------------------------------------------------------
    if 'arrow' in kwargs:
        if not isinstance(kwargs['arrow'], tuple):
            print("Arrow must be a tuple (X,Y)")
            show_arrow = False
        else:
            coor_arrow = kwargs['arrow']
            show_arrow = True
    else:
        show_arrow = False

    if 'arrowshiftx' in kwargs:
        arrow_shiftx = kwargs['arrowshiftx']
    else:
        arrow_shiftx = 5

    if 'arrowlength' in kwargs:
        arrow_length = kwargs['arrowlength']
    else:
        arrow_length = 20

    if 'arrowalpha' in kwargs:
        arrow_alpha = kwargs['arrowalpha']
    else:
        arrow_alpha = 0.8

    # LABEL --------------------------------------------------------------------
    if 'label' in kwargs:
        label = kwargs['label']
        if len(label) != num_plots:
            print("Label list does not have enough items")
            label = None
    else:
        label = None

    if 'labelsize' in kwargs:
        labelsize = kwargs['labelsize']
    else:
        labelsize = 12

    if 'labelpad' in kwargs:
        labelpad = kwargs['labelpad']
    else:
        labelpad = 5

    # GRID ---------------------------------------------------------------------
    if 'grid' in kwargs:
        grid = kwargs['grid']
    else:
        grid = False

    if 'gridcolor' in kwargs:
        grid_color = kwargs['gridcolor']
    else:
        grid_color = '#f7f7f7'

    if 'gridspacing' in kwargs:
        grid_spacing = kwargs['gridspacing']
    else:
        grid_spacing = 10

    if 'gridalpha' in kwargs:
        grid_alpha = kwargs['gridalpha']
    else:
        grid_alpha = 0.3

    # VMAX-VMIN ----------------------------------------------------------------
    if 'vmax' in kwargs:
        if isinstance(kwargs['vmax'], tuple) or isinstance(kwargs['vmax'],
                                                           list):
            if len(kwargs['vmax']) != num_plots:
                print("Vmax list does not have enough items, setting all to None")
                vmax = [None for i in range(num_plots)]
            else:
                vmax = kwargs['vmax']
        else:
            vmax = [kwargs['vmax'] for i in range(num_plots)]
    else:
        vmax = [None for i in range(num_plots)]

    if 'vmin' in kwargs:
        if isinstance(kwargs['vmin'], tuple) or isinstance(kwargs['vmin'],
                                                           list):
            if len(kwargs['vmin']) != num_plots:
                print(
                "Vmax list does not have enough items, setting all to None")
                vmin = [None for i in range(num_plots)]
            else:
                vmin = kwargs['vmin']
        else:
            vmin = [kwargs['vmin'] for i in range(num_plots)]
    else:
        vmin = [None for i in range(num_plots)]

    # CROSS --------------------------------------------------------------------
    if 'cross' in kwargs:
        if not isinstance(kwargs['cross'], tuple):
            print("Crosshair must be a tuple (X,Y)")
            show_cross = False
        else:
            coor_cross = kwargs['cross']
            show_cross = True
    else:
        show_cross = False
    if 'crossalpha' in kwargs: cross_alpha = kwargs['crossalpha']
    else:
        cross_alpha = 0.4

    # AXIS - ANGSCALE ----------------------------------------------------------
    if 'angscale' in kwargs:
        angscale = kwargs['angscale']
    else:
        angscale = False
    if 'angticksep' in kwargs:
        angticksep = kwargs['angticksep']
    else:
        angticksep = 50
    if 'pxscale' in kwargs:
        pxscale = kwargs['pxscale']
    else:
        pxscale = 0.01  # default for Keck/NIRC2
    if 'axis' in kwargs:
        show_axis = kwargs['axis']
    else:
        show_axis = True
    # --------------------------------------------------------------------------

    if 'showcent' in kwargs:
        show_center = True
    else:
        show_center = False

    if 'getfig' in kwargs:
        getfig = kwargs['getfig']
    else:
        getfig = False

    if 'save' in kwargs and isinstance(kwargs['save'], str):
        save = True
        savepath = kwargs['save']
    else:
        save = False

    # Defaults previously used: 'magma','CMRmap','RdBu_r'
    if 'cmap' in kwargs:
        custom_cmap = kwargs['cmap']
        if not isinstance(custom_cmap, list):
            custom_cmap = [kwargs['cmap'] for i in range(num_plots)]
        else:
            if not len(custom_cmap) == num_plots:
                raise RuntimeError('Cmap list does not have enough items')
    else:
        custom_cmap = ['viridis' for i in range(num_plots)]

    if 'log' in kwargs and kwargs['log'] is True:
        logscale = kwargs['log']
    else:
        logscale = False

    if 'colorb' in kwargs:
        colorb = kwargs['colorb']
    else:
        colorb = True
    
    if 'dpi' in kwargs:
        dpi = kwargs['dpi']
    else:
        dpi = 90

    if 'title' in kwargs:
        tit = kwargs['title']
    else:
        tit = None
    
    if 'horsp' in kwargs:
        hor_spacing = kwargs['horsp']
    else:
        hor_spacing = 0.3
    
    if 'versp' in kwargs:
        ver_spacing = kwargs['versp']
    else:
        ver_spacing = 0.2

    # --------------------------------------------------------------------------

    subplot_size = 4
    if rows == 0:
        raise ValueError('Rows must be a positive integer')
    fig = figure(figsize=(cols*subplot_size, rows*subplot_size), dpi=dpi)
    
    if tit is not None:
        fig.suptitle(tit, fontsize=14)
    
    for i, v in enumerate(range(num_plots)):
        image = data[i].copy()
        frame_size = image.shape[0]  # assuming square frames
        cy, cx = frame_center(image)

        v += 1
        ax = subplot(rows, cols, v)
        ax.set_aspect('equal')
        if logscale:
            image += np.abs(image.min())
            norm = colors.LogNorm(vmin=max(image.min(), 0.1), vmax=image.max())
        else:
            norm = None
        im = ax.imshow(image, cmap=custom_cmap[i], interpolation='nearest',
                       origin='lower', vmin=vmin[i], vmax=vmax[i],
                       norm=norm)

        if show_circle:
            for j in range(len(coor_circle)):
                circle = Circle((coor_circle[j][0], coor_circle[j][1]),
                                radius=circle_rad, color='white', fill=False,
                                alpha=circle_alpha)
                ax.add_artist(circle)

        if show_cross:
            ax.scatter([coor_cross[0]], [coor_cross[1]], marker='+',
                       color='white', alpha=cross_alpha)

        if show_center:
            ax.axhline(cx, xmin=0, xmax=frame_size, alpha=0.5,
                       linestyle='dashed', color='white', lw=0.8)
            ax.axvline(cy, ymin=0, ymax=frame_size, alpha=0.5,
                       linestyle='dashed', color='white', lw=0.8)

        if show_arrow:
            ax.arrow(coor_arrow[0]+arrow_length+arrow_shiftx, coor_arrow[1],
                     -arrow_length, 0, color='white', head_width=10,
                     head_length=8, width=3, length_includes_head=True,
                     alpha=arrow_alpha)
            
        if label is not None:
            ax.annotate(label[i], xy=(labelpad,labelpad), color='white',
                        xycoords='axes pixels', weight='bold', size=labelsize)    

        if colorb:
            # create an axes on the right side of ax. The width of cax is 5%
            # of ax and the padding between cax and ax wis fixed at 0.05 inch
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            cb = colorbar(im, ax=ax, cax=cax, drawedges=False)
            cb.outline.set_linewidth(0.1)
            cb.ax.tick_params(labelsize=8) 

        if grid:
            ax.tick_params(axis='both', which='minor')
            minor_ticks = np.arange(0, data[i].shape[0], grid_spacing)
            ax.set_xticks(minor_ticks, minor=True)
            ax.set_yticks(minor_ticks, minor=True)
            ax.grid('on', which='minor', color=grid_color, linewidth=1,
                    alpha=grid_alpha)
        else:
            ax.grid('off')

        if angscale:
            # Converting axes from pixels to arcseconds
            half_num_ticks = int(np.round(cy // angticksep))

            # Calculate the pixel locations at which to put ticks
            ticks = []
            for t in range(half_num_ticks, -half_num_ticks-1, -1):
                # Avoid ticks not showing on the last pixel
                if not cy - t * angticksep == frame_size:
                    ticks.append(cy - t * angticksep)
                else:
                    ticks.append((cy - t * angticksep) - 1)
            ax.set_xticks(ticks)
            ax.set_yticks(ticks)

            # Corresponding distance in arcseconds, measured from the center
            labels = []
            for t in range(half_num_ticks, -half_num_ticks-1, -1):
                labels.append(0.0 - t * (angticksep * pxscale))
                #print xlabels
            ax.set_xticklabels(labels)
            ax.set_yticklabels(labels)
            ax.set_xlabel("arcseconds", fontsize=12)
            ax.set_ylabel("arcseconds", fontsize=12)
            ax.tick_params(axis='both', which='major', labelsize=10)

        if not show_axis:
            ax.set_axis_off()
    
    fig.subplots_adjust(wspace=hor_spacing, hspace=ver_spacing)
    if save:
        savefig(savepath, dpi=dpi, bbox_inches='tight')

    if getfig:
        return fig
    else:
        show()


def plot_surface(image, center=None, size=15, output=False, ds9_indexing=False, 
                 **kwargs):
    """
    Create a surface plot from image.
    
    By default, the whole image is plotted. The 'center' and 'size' attributes 
    allow to crop the image.
        
    Parameters
    ----------
    image : numpy.array
        The image as a numpy.array.
    center : tuple of 2 int (optional, default=None)
        If None, the whole image will be plotted. Otherwise, it grabs a square
        subimage at the 'center' (Y,X) from the image.
    size : int (optional, default=15)
        It corresponds to the size of a square in the image.
    output : {False, True}, bool optional
        Whether to output the grids and intensities or not.
    ds9_indexing : {False, True}, bool optional 
        If True the coordinates are in X,Y convention and in 1-indexed format.
    kwargs:
        Additional attributes are passed to the matplotlib figure() and 
        plot_surface() method.        
    
    Returns
    -------
    out : tuple of 3 numpy.array
        x and y for the grid, and the intensity
        
    """        
    if center is not None: 
        if ds9_indexing:
            center = (center[0]-1,center[1]-1) 
            cx, cy = center
        else:
            cy, cx = center
        if size % 2:            # if size is odd             
            x = np.outer(np.arange(0,size,1), np.ones(size))
        else:                   # otherwise, size is even
            x = np.outer(np.arange(0,size+1,1), np.ones(size+1))
        y = x.copy().T            
        z = image[cy-size//2:cy+size//2+1,cx-size//2:cx+size//2+1]   
    else:
        cy, cx = frame_center(image)
        if size is not None:
            if size % 2:
                x = np.outer(np.arange(0,size,1), np.ones(size))
            else: 
                x = np.outer(np.arange(0,size+1,1), np.ones(size+1))
            y = x.copy().T
            z = image[cy-size//2:cy+size//2+1,cx-size//2:cx+size//2+1]
        else:
            size = image.shape[0]
            x = np.outer(np.arange(0,size,1), np.ones(size))
            y = x.copy().T 
            z = image        
    
    figure(figsize=kwargs.pop('figsize',(6,6)))
    ax = axes(projection='3d')
    ax.dist = 12
    ax.plot_surface(x, y, z, rstride=1, cstride=1, linewidth=0, **kwargs) 
    ax.set_xlabel('$x$')
    ax.set_ylabel('$y$')
    ax.set_zlabel('$I(x,y)$')
    ax.xaxis._axinfo['label']['space_factor'] = 2
    ax.yaxis._axinfo['label']['space_factor'] = 2
    ax.zaxis._axinfo['label']['space_factor'] = 3.5
    #ax.zaxis._axinfo['ticklabel']['space_factor'] = 1.5
    ax.set_title('Data')
    show()
    
    if output:
        return (x,y,z)
 

def lines_of_code():
    """ Calculates the lines of code for VIP pipeline. Not objective measure 
    of developer's work! (note to self). 
    """
    cur_path = os.path.dirname(os.path.abspath(__file__))
    path = cur_path[:-len('var')]

    ignore_set = set(["__init__.py"])
    
    loclist = []

    for pydir, _, pyfiles in os.walk(path):
        if 'exlib/' not in pydir:
            for pyfile in pyfiles:
                if pyfile not in ignore_set and pyfile.endswith(".py"):
                    totalpath = os.path.join(pydir, pyfile)
                    loclist.append((len(open(totalpath,"r").read().splitlines()),
                                       totalpath.split(path)[1]) )

    for linenumbercount, filename in loclist: 
        print("%05d lines in %s" % (linenumbercount, filename))

    msg = "\nTotal: {:} lines in ({:}) excluding external libraries."
    print(msg.format(sum([x[0] for x in loclist]), path))




 



    



#! /usr/bin/env python

"""
"""

from __future__ import division

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['thresholding']

import numpy as np


def thresholding(array, threshold, mode):
    """ Array thresholding strategies.
    """
    x = array.copy()
    if mode == 'soft':
        j = np.abs(x) <= threshold
        x[j] = 0
        k = np.abs(x) > threshold
        if isinstance(threshold, float):
            x[k] = x[k] - np.sign(x[k]) * threshold
        else:
            x[k] = x[k] - np.sign(x[k]) * threshold[k]
    elif mode == 'hard':
        j = np.abs(x) < threshold
        x[j] = 0
    elif mode == 'nng':
        j = np.abs(x) <= threshold
        x[j] = 0
        j = np.abs(x) > threshold
        x[j] = x[j] - threshold**2/x[j]
    elif mode == 'greater':
        j = x < threshold
        x[j] = 0
    elif mode == 'less':
        j = x > threshold
        x[j] = 0
    else:
        raise RuntimeError('Thresholding mode not recognized')
    return x


"""
Subpackage ``stats`` contains functionalities such as:
    - extracting statistics (mean, median, std dev, sum) in regions of a frame
    or cube,
    - median absolute deviation,
    - sigma filtering of pixels in frames,
    - distance (correlation) between the frames in a cube,
    - distance (correlation) between a cube and a reference frame.
"""
from __future__ import absolute_import

from .clip_sigma import *
from .cube_stats import *
from .distances import *
from .im_stats import *
from .utils_stats import *



#! /usr/bin/env python

"""
Module with helping functions.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = []

import numpy as np
from ..var import get_square_robust, frame_center, prepare_matrix
from ..preproc import cube_derotate, cube_collapse, cube_rescaling
from .svd import svd_wrapper


def scale_cube_for_pca(cube,scal_list, full_output=True, inverse=False, y_in=1,
                       x_in=1, imlib='opencv', interpolation='lanczos4'):
    """
    Wrapper to scale or descale a cube by factors given in scal_list, without 
    any loss of information (zero-padding if scaling > 1).
    Important: in case of ifs data, the scaling factors in var_list should be 
    >= 1 (ie. provide the scaling factors as for scaling to the longest 
    wavelength channel).

    Parameters:
    -----------
    cube: 3D-array
       Datacube that whose frames have to be rescaled.
    scal_list: 1D-array
       Vector of same dimension as the first dimension of datacube, containing 
       the scaling factor for each frame.
    full_output: bool, {True,False}, optional
       Whether to output just the rescaled cube (False) or also its median, 
       the new y and x shapes of the cube, and the new centers cy and cx of the 
       frames (True).
    inverse: bool, {True,False}, optional
       Whether to inverse the scaling factors in scal_list before applying them 
       or not; i.e. True is to descale the cube (typically after a first scaling
       has already been done)
    y_in, x-in:
       Initial y and x sizes. In case the cube is descaled, these values will
       be used to crop back the cubes/frames to their original size.
    imlib : str optional
        See the documentation of the ``vip_hci.preproc.frame_rescaling``
        function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rescaling``
        function.

    Returns:
    --------
    frame: 2D-array
        The median of the rescaled cube.
    If full_output is set to True, the function returns:
    cube,frame,y,x,cy,cx: 3D-array,2D-array,int,int,int,int
        The rescaled cube, its median, the new y and x shapes of the cube, and 
        the new centers cy and cx of the frames
    """
    # Cube zeros-padding to avoid loosing info when scaling the cube
    # TODO: pad with random gaussian noise instead of zeros. Padding with
    # only zeros can make the svd not converge in a pca per zone.

    n, y, x = cube.shape

    max_sc = np.amax(scal_list)

    if not inverse and max_sc > 1:
        new_y = int(np.ceil(max_sc*y))
        new_x = int(np.ceil(max_sc*x))
        if (new_y - y) % 2 != 0:
            new_y = new_y+1
        if (new_x - x) % 2 != 0:
            new_x = new_x+1
        pad_len_y = (new_y - y)//2
        pad_len_x = (new_x - x)//2
        big_cube = np.pad(cube, ((0,0), (pad_len_y, pad_len_y), 
                                 (pad_len_x, pad_len_x)), 'constant', 
                          constant_values=(0,))
    else: 
        big_cube = cube.copy()

    n, y, x = big_cube.shape
    cy, cx = frame_center(big_cube[0])
    var_list = scal_list

    if inverse:
        var_list = 1./scal_list[:]
        cy, cx = frame_center(cube[0])

    # (de)scale the cube, so that a planet would now move radially
    cube, frame = cube_rescaling(big_cube, var_list, ref_y=cy, ref_x=cx,
                                 imlib=imlib, interpolation=interpolation)

    if inverse:
        if max_sc > 1:
            # TODO: check the use of get_square_robust
            frame = get_square_robust(frame,max(y_in,x_in), cy, cx,strict=False)
            if full_output:
                n_z = cube.shape[0]
                array_old = cube.copy()
                cube = np.zeros([n_z,max(y_in, x_in), max(y_in, x_in)])
                for zz in range(n_z):
                    # TODO: check the use of get_square_robust
                    cube[zz] = get_square_robust(array_old[zz],max(y_in, x_in),
                                                 cy, cx, strict=False)

    if full_output: 
        return cube, frame, y, x, cy, cx
    else: 
        return frame


def pca_annulus(cube, angs, ncomp, annulus_width, r_guess, cube_ref=None,
                svd_mode='lapack', scaling=None, collapse='median',
                imlib='opencv', interpolation='lanczos4'):
    """
    PCA process the cube only for an annulus of a given width and at a given
    radial distance to the frame center. It returns a PCA processed frame with 
    only non-zero values at the positions of the annulus.
    
    Parameters
    ----------
    cube : array_like
        The cube of fits images expressed as a numpy.array.
    angs : array_like
        The parallactic angles expressed as a numpy.array.
    ncomp : int
        The number of principal component.
    annulus_width : float
        The annulus width in pixel on which the PCA is performed.
    r_guess : float
        Radius of the annulus in pixels.
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With
        "temp-mean" then temporal px-wise mean subtraction is done, with
        "spat-mean" then the spatial mean is subtracted, with "temp-standard"
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is returned.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    
    Returns
    -------
    Depending on ``collapse`` parameter a final collapsed frame or the cube of
    residuals is returned.
    """
    data, ind = prepare_matrix(cube, scaling, mode='annular',
                               annulus_radius=r_guess, verbose=False,
                               annulus_width=annulus_width)
    yy, xx = ind

    if cube_ref is not None:
        data_svd, _ = prepare_matrix(cube_ref, scaling, mode='annular',
                                     annulus_radius=r_guess, verbose=False,
                                     annulus_width=annulus_width)
    else:
        data_svd = data
        
    V = svd_wrapper(data_svd, svd_mode, ncomp, debug=False, verbose=False)
        
    transformed = np.dot(data, V.T)
    reconstructed = np.dot(transformed, V)                           
    residuals = data - reconstructed
    cube_zeros = np.zeros_like(cube)
    cube_zeros[:, yy, xx] = residuals

    if angs is not None:
        cube_res_der = cube_derotate(cube_zeros, angs, imlib=imlib,
                                     interpolation=interpolation)
        if collapse is not None:
            pca_frame = cube_collapse(cube_res_der, mode=collapse)
            return pca_frame
        else:
            return cube_res_der

    else:
        if collapse is not None:
            pca_frame = cube_collapse(cube_zeros, mode=collapse)
            return pca_frame
        else:
            return cube_zeros







#! /usr/bin/env python

"""
Module with functions for outlier frame detection.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['cube_detect_badfr_pxstats',
           'cube_detect_badfr_ellipticipy',
           'cube_detect_badfr_correlation']

import numpy as np
import pandas as pn
from matplotlib import pyplot as plt
from photutils import detection
from astropy.stats import sigma_clip
from ..var import get_annulus
from ..conf import time_ini, timing
from ..stats import (cube_stats_aperture, cube_stats_annulus, 
                              cube_distance)


def cube_detect_badfr_pxstats(array, mode='annulus', in_radius=10, width=10, 
                              top_sigma=1.0, low_sigma=1.0, window=None, 
                              plot=True, verbose=True):             
    """ Returns the list of bad frames from a cube using the px statistics in 
    a centered annulus or circular aperture. Frames that are more than a few 
    standard deviations discrepant are rejected. Should be applied on a 
    recentered cube.
    
    Parameters
    ----------
    array : array_like 
        Input 3d array, cube.
    mode : {'annulus', 'circle'}, string optional
        Whether to take the statistics from a circle or an annulus.
    in_radius : int optional
        If mode is 'annulus' then 'in_radius' is the inner radius of the annular 
        region. If mode is 'circle' then 'in_radius' is the radius of the 
        aperture.
    width : int optional
        Size of the annulus. Ignored if mode is 'circle'.
    top_sigma : int, optional
        Top boundary for rejection.
    low_sigma : int, optional
        Lower boundary for rejection.
    window : int, optional
        Window for smoothing the median and getting the rejection statistic.
    plot : {True, False}, bool optional
        If true it plots the mean fluctuation as a function of the frames and 
        the boundaries.
    verbose : {True, False}, bool optional
        Whether to print to stdout or not.
    
    Returns
    -------
    good_index_list : array_like
        1d array of good indices.
    bad_index_list : array_like
        1d array of bad frames indices.
    
    """
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')
    if in_radius+width > array[0].shape[0]/2.:
        msgve = 'Inner radius and annulus size are too big (out of boundaries)'
        raise ValueError(msgve)
    
    if verbose:  start_time = time_ini()
    
    n = array.shape[0]
    
    if mode=='annulus':
        mean_values,_,_,_ = cube_stats_annulus(array, inner_radius=in_radius, 
                                               size=width, full_out=True)
    elif mode=='circle':
        _,mean_values,_,_ = cube_stats_aperture(array, radius=in_radius, 
                                                full_output=True)
    else: 
        raise TypeError('Mode not recognized')
    
    if window is None:  window = int(n/3.)
    mean_smooth = pn.rolling_median(mean_values, window , center=True)
    temp = pn.Series(mean_smooth)
    temp = temp.fillna(method='backfill')
    temp = temp.fillna(method='ffill')
    mean_smooth = temp.values
    sigma = np.std(mean_values)
    bad_index_list = []
    good_index_list = []
    top_boundary = np.empty([n])
    bot_boundary = np.empty([n])
    for i in range(n):
        if mode=='annulus':
            i_mean_value = get_annulus(array[i], inner_radius=in_radius, 
                                        width=width, output_values=True).mean()
        elif mode=='circle':
            i_mean_value = mean_values[i]
        top_boundary[i] = mean_smooth[i] + top_sigma*sigma
        bot_boundary[i] = mean_smooth[i] - low_sigma*sigma
        if (i_mean_value > top_boundary[i] or i_mean_value < bot_boundary[i]):
            bad_index_list.append(i)
        else:
            good_index_list.append(i)                       

    if verbose:
        bad = len(bad_index_list)
        percent_bad_frames = (bad*100)/n
        msg1 = "Done detecting bad frames from cube: {:} out of {:} ({:.3}%)"
        print(msg1.format(bad, n, percent_bad_frames)) 

    if plot:
        plt.figure(figsize=(10, 6), dpi=100)
        plt.plot(mean_values, 'o', label='mean fluctuation', lw = 1.4)
        plt.plot(mean_smooth, label='smoothed median', lw = 2, ls='-', alpha=0.5)
        plt.plot(top_boundary, label='top limit', lw = 1.4, ls='-')
        plt.plot(bot_boundary, label='lower limit', lw = 1.4, ls='-')
        plt.legend(fancybox=True, framealpha=0.5)

    if verbose:  timing(start_time)

    good_index_list = np.array(good_index_list)
    bad_index_list = np.array(bad_index_list)
    
    return good_index_list, bad_index_list


def cube_detect_badfr_ellipticipy(array, fwhm, roundlo=-0.2, roundhi=0.2,
                                  verbose=True):
    """ Returns the list of bad frames  from a cube by measuring the PSF 
    ellipticity of the central source. Should be applied on a recentered cube.
    
    Parameters
    ----------
    array : array_like 
        Input 3d array, cube.
    fwhm : float
        FWHM size in pixels.
    roundlo, roundhi : float, optional
        Lower and higher bounds for the ellipticipy.
    verbose : {True, False}, bool optional
        Whether to print to stdout or not.
        
    Returns
    -------
    good_index_list : array_like
        1d array of good indices.
    bad_index_list : array_like
        1d array of bad frames indices.
    
    Notes
    -----
    From photutils.daofind documentation:
    DAOFIND calculates the object roundness using two methods.  The 'roundlo' 
    and 'roundhi' bounds are applied to both measures of roundness.  The first 
    method ('roundness1'; called 'SROUND' in DAOFIND) is based on the source 
    symmetry and is the ratio of a measure of the object's bilateral (2-fold) 
    to four-fold symmetry. The second roundness statistic ('roundness2'; called 
    'GROUND' in DAOFIND) measures the ratio of the difference in the height of
    the best fitting Gaussian function in x minus the best fitting Gaussian 
    function in y, divided by the average of the best fitting Gaussian 
    functions in x and y.  A circular source will have a zero roundness. A 
    source extended in x or y will have a negative or positive roundness, 
    respectively.
    
    """
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')
    
    if verbose:  start_time = time_ini()
    
    n = array.shape[0]
    
    # Calculate a 2D Gaussian density enhancement kernel
    gauss2d_kernel = detection.findstars._FindObjKernel
    # Find sources in an image by convolving the image with the input kernel 
    # and selecting connected pixels above a given threshold
    find_objs = detection.findstars._findobjs
    # Find the properties of each detected source, as defined by DAOFIND
    obj_prop = detection.findstars._daofind_properties

    ff_clipped = sigma_clip(array[0], sig=4, iters=None)
    thr = ff_clipped.max()
    
    goodfr = []
    badfr = []
    for i in range(n):
        # we create a circular gaussian kernel
        kernel = gauss2d_kernel(fwhm=fwhm, ratio=1.0, theta=0.0)  
        objs = find_objs(array[i], threshold=thr, kernel=kernel, 
                         exclude_border=True)
        tbl = obj_prop(objs, threshold=thr, kernel=kernel, sky=0)
        # we mask the peak px object
        table_mask = (tbl['peak'] == tbl['peak'].max())
        tbl = tbl[table_mask]
        roun1 = tbl['roundness1'][0]
        roun2 = tbl['roundness2'][0]
        # we check the roundness
        if roun1>roundlo and roun1<roundhi and roun2>roundlo and roun2<roundhi:  
            goodfr.append(i)
        else:
            badfr.append(i)
    
    bad_index_list = np.array(badfr)
    good_index_list = np.array(goodfr)
    if verbose:
        bad = len(bad_index_list)
        percent_bad_frames = (bad*100)/n
        msg1 = "Done detecting bad frames from cube: {:} out of {:} ({:.3}%)"
        print(msg1.format(bad, n, percent_bad_frames)) 
    
    if verbose:  timing(start_time)
    
    return good_index_list, bad_index_list


def cube_detect_badfr_correlation(array, frame_ref, crop_size=30, dist='pearson',
                                  percentile=20, plot=True, verbose=True):
    """ Returns the list of bad frames from a cube by measuring the distance 
    (similarity) or correlation of the frames (cropped to a 30x30 subframe) 
    wrt a reference frame from the same cube. Then the distance/correlation 
    level is thresholded (percentile parameter) to find the outliers. Should be 
    applied on a recentered cube.
    
    Parameters
    ----------
    array : array_like 
        Input 3d array, cube.
    frame_ref : int
        Index of the frame that will be used as a reference. Must be of course
        a frame taken with a good wavefront quality.
    dist : {'sad','euclidean','mse','pearson','spearman'}, str optional
        One of the similarity or disimilarity measures from function 
        vip_hci.stats.distances.cube_distance(). 
    percentile : int
        The percentage of frames that will be discarded. 
    plot : {True, False}, bool optional
        If true it plots the mean fluctuation as a function of the frames and 
        the boundaries.
    verbose : {True, False}, bool optional
        Whether to print to stdout or not.
            
    Returns
    -------
    good_index_list : array_like
        1d array of good indices.
    bad_index_list : array_like
        1d array of bad frames indices.
        
    """
    from .cosmetics import cube_crop_frames
    
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')
    
    if verbose:  start_time = time_ini()
    
    n = array.shape[0]
    # the cube is cropped to the central area
    subarray = cube_crop_frames(array, min(crop_size, array.shape[1]), verbose=False)
    distances = cube_distance(subarray, frame_ref, 'full', dist, plot=False)
        
    if dist=='pearson' or dist=='spearman': # measures of correlation or similarity
        minval = np.min(distances[~np.isnan(distances)])
        distances = np.nan_to_num(distances)
        distances[np.where(distances==0)] = minval
        threshold = np.percentile(distances, percentile)
        indbad = np.where(distances <= threshold)
        indgood = np.where(distances > threshold)
    else:                                                   # measures of dissimilarity
        threshold = np.percentile(distances, 100-percentile)
        indbad = np.where(distances >= threshold)
        indgood = np.where(distances < threshold)
        
    bad_index_list = indbad[0]
    good_index_list = indgood[0]
    if verbose:
        bad = len(bad_index_list)
        percent_bad_frames = (bad*100)/n
        msg1 = "Done detecting bad frames from cube: {:} out of {:} ({:.3}%)"
        print(msg1.format(bad, n, percent_bad_frames)) 
    
    if plot:
        lista = distances
        _, ax = plt.subplots(figsize=(10, 6), dpi=100)
        x = range(len(lista))
        ax.plot(x, lista, '-', color='blue', alpha=0.3)
        if n>5000:
            ax.plot(x, lista, ',', color='blue', alpha=0.5)
        else:
            ax.plot(x, lista, '.', color='blue', alpha=0.5)
        ax.vlines(frame_ref, ymin=np.nanmin(lista), ymax=np.nanmax(lista), 
                   colors='green', linestyles='dashed', lw=2, alpha=0.8,
                   label='Reference frame '+str(frame_ref))
        ax.hlines(np.median(lista), xmin=-1, xmax=n+1, colors='purple', 
                   linestyles='solid', label='Median value')
        ax.hlines(threshold, xmin=-1, xmax=n+1, lw=2, colors='red', 
                   linestyles='dashed', label='Threshold')
        plt.xlabel('Frame number')
        if dist=='sad':
            plt.ylabel('SAD - Manhattan distance')
        elif dist=='euclidean':
            plt.ylabel('Euclidean distance')
        elif dist=='pearson':
            plt.ylabel('Pearson correlation coefficient')
        elif dist=='spearman':
            plt.ylabel('Spearman correlation coefficient')
        elif dist=='mse':
            plt.ylabel('Mean squared error')
        elif dist=='ssim':
            plt.ylabel('Structural Similarity Index')
        
        plt.xlim(xmin=-1, xmax=n+1)
        plt.minorticks_on()
        plt.legend(fancybox=True, framealpha=0.5, fontsize=12, loc='best')
        plt.grid(which='mayor')
    
    if verbose:  timing(start_time)
    
    return good_index_list, bad_index_list





"""
Code with the Isotropic Undecimated Wavelet Transform, taken (Aug 24, 2015) from 
https://github.com/ratt-ru/PyMORESANE/
Credits to J. S. Kenyon
"""

import numpy as np
import multiprocessing as mp
import ctypes


def iuwt_decomposition(in1, scale_count, scale_adjust=0, mode='ser', core_count=2, store_smoothed=False):
    """
    This function serves as a handler for the different implementations of the IUWT decomposition. It allows the
    different methods to be used almost interchangeably.

    INPUTS:
    in1                 (no default):       Array on which the decomposition is to be performed.
    scale_count         (no default):       Maximum scale to be considered.
    scale_adjust        (default=0):        Adjustment to scale value if first scales are of no interest.
    mode                (default='ser'):    Implementation of the IUWT to be used - 'ser', 'mp'.
    core_count          (default=1):        Additional option for multiprocessing - specifies core count.
    store_smoothed      (default=False):    Boolean specifier for whether the smoothed image is stored or not.

    OUTPUTS:
    Returns the decomposition with the additional smoothed coefficients if specified.
    """

    if mode=='ser':
        return ser_iuwt_decomposition(in1, scale_count, scale_adjust, store_smoothed)
    elif mode=='mp':
        return mp_iuwt_decomposition(in1, scale_count, scale_adjust, store_smoothed, core_count)


def iuwt_recomposition(in1, scale_adjust=0, mode='ser', core_count=1, store_on_gpu=False, smoothed_array=None):
    """
    This function serves as a handler for the different implementations of the IUWT recomposition. It allows the
    different methods to be used almost interchangeably.

    INPUTS:
    in1                 (no default):       Array on which the decomposition is to be performed.
    scale_adjust        (no default):       Number of omitted scales.
    mode                (default='ser')     Implementation of the IUWT to be used - 'ser', 'mp' or 'gpu'.
    core_count          (default=1)         Additional option for multiprocessing - specifies core count.
    store_on_gpu        (default=False):    Boolean specifier for whether the decomposition is stored on the gpu or not.

    OUTPUTS:
    Returns the recomposition.
    """

    if mode=='ser':
        return ser_iuwt_recomposition(in1, scale_adjust, smoothed_array)
    elif mode=='mp':
        return mp_iuwt_recomposition(in1, scale_adjust, core_count, smoothed_array)


def ser_iuwt_decomposition(in1, scale_count, scale_adjust, store_smoothed):
    """
    This function calls the a trous algorithm code to decompose the input into its wavelet coefficients. This is
    the isotropic undecimated wavelet transform implemented for a single CPU core.

    INPUTS:
    in1                 (no default):   Array on which the decomposition is to be performed.
    scale_count         (no default):   Maximum scale to be considered.
    scale_adjust        (default=0):    Adjustment to scale value if first scales are of no interest.
    store_smoothed      (default=False):Boolean specifier for whether the smoothed image is stored or not.

    OUTPUTS:
    detail_coeffs                       Array containing the detail coefficients.
    C0                  (optional):     Array containing the smoothest version of the input.
    """

    wavelet_filter = (1./16)*np.array([1,4,6,4,1])      # Filter-bank for use in the a trous algorithm.

    # Initialises an empty array to store the coefficients.

    detail_coeffs = np.empty([scale_count-scale_adjust, in1.shape[0], in1.shape[1]])

    C0 = in1    # Sets the initial value to be the input array.

    # The following loop, which iterates up to scale_adjust, applies the a trous algorithm to the scales which are
    # considered insignificant. This is important as each set of wavelet coefficients depends on the last smoothed
    # version of the input.

    if scale_adjust>0:
        for i in range(0, scale_adjust):
            C0 = ser_a_trous(C0, wavelet_filter, i)

    # The meat of the algorithm - two sequential applications fo the a trous followed by determination and storing of
    # the detail coefficients. C0 is reassigned the value of C on each loop - C0 is always the smoothest version of the
    # input image.

    for i in range(scale_adjust,scale_count):
        C = ser_a_trous(C0, wavelet_filter, i)                                  # Approximation coefficients.
        C1 = ser_a_trous(C, wavelet_filter, i)                                  # Approximation coefficients.
        detail_coeffs[i-scale_adjust,:,:] = C0 - C1                             # Detail coefficients.
        C0 = C

    if store_smoothed:
        return detail_coeffs, C0
    else:
        return detail_coeffs

def ser_iuwt_recomposition(in1, scale_adjust, smoothed_array):
    """
    This function calls the a trous algorithm code to recompose the input into a single array. This is the
    implementation of the isotropic undecimated wavelet transform recomposition for a single CPU core.

    INPUTS:
    in1             (no default):   Array containing wavelet coefficients.
    scale_adjust    (no default):   Indicates the number of truncated array pages.
    smoothed_array  (default=None): For a complete inverse transform, this must be the smoothest approximation.

    OUTPUTS:
    recomposition                   Array containing the reconstructed image.
    """

    wavelet_filter = (1./16)*np.array([1,4,6,4,1])      # Filter-bank for use in the a trous algorithm.

    # Determines scale with adjustment and creates a zero array to store the output, unless smoothed_array is given.

    max_scale = in1.shape[0] + scale_adjust

    if smoothed_array is None:
        recomposition = np.zeros([in1.shape[1], in1.shape[2]])
    else:
        recomposition = smoothed_array

    # The following loops call the a trous algorithm code to recompose the input. The first loop assumes that there are
    # non-zero wavelet coefficients at scales above scale_adjust, while the second loop completes the recomposition
    # on the scales less than scale_adjust.

    for i in range(max_scale-1, scale_adjust-1, -1):
        recomposition = ser_a_trous(recomposition, wavelet_filter, i) + in1[i-scale_adjust,:,:]

    if scale_adjust>0:
        for i in range(scale_adjust-1, -1, -1):
            recomposition = ser_a_trous(recomposition, wavelet_filter, i)

    return recomposition

def ser_a_trous(C0, filter, scale):
    """
    The following is a serial implementation of the a trous algorithm. Accepts the following parameters:

    INPUTS:
    filter      (no default):   The filter-bank which is applied to the components of the transform.
    C0          (no default):   The current array on which filtering is to be performed.
    scale       (no default):   The scale for which the decomposition is being carried out.

    OUTPUTS:
    C1                          The result of applying the a trous algorithm to the input.
    """
    tmp = filter[2]*C0

    tmp[(2**(scale+1)):,:] += filter[0]*C0[:-(2**(scale+1)),:]
    tmp[:(2**(scale+1)),:] += filter[0]*C0[(2**(scale+1))-1::-1,:]

    tmp[(2**scale):,:] += filter[1]*C0[:-(2**scale),:]
    tmp[:(2**scale),:] += filter[1]*C0[(2**scale)-1::-1,:]

    tmp[:-(2**scale),:] += filter[3]*C0[(2**scale):,:]
    tmp[-(2**scale):,:] += filter[3]*C0[:-(2**scale)-1:-1,:]

    tmp[:-(2**(scale+1)),:] += filter[4]*C0[(2**(scale+1)):,:]
    tmp[-(2**(scale+1)):,:] += filter[4]*C0[:-(2**(scale+1))-1:-1,:]

    C1 = filter[2]*tmp

    C1[:,(2**(scale+1)):] += filter[0]*tmp[:,:-(2**(scale+1))]
    C1[:,:(2**(scale+1))] += filter[0]*tmp[:,(2**(scale+1))-1::-1]

    C1[:,(2**scale):] += filter[1]*tmp[:,:-(2**scale)]
    C1[:,:(2**scale)] += filter[1]*tmp[:,(2**scale)-1::-1]

    C1[:,:-(2**scale)] += filter[3]*tmp[:,(2**scale):]
    C1[:,-(2**scale):] += filter[3]*tmp[:,:-(2**scale)-1:-1]

    C1[:,:-(2**(scale+1))] += filter[4]*tmp[:,(2**(scale+1)):]
    C1[:,-(2**(scale+1)):] += filter[4]*tmp[:,:-(2**(scale+1))-1:-1]

    return C1

def mp_iuwt_decomposition(in1, scale_count, scale_adjust, store_smoothed, core_count):
    """
    This function calls the a trous algorithm code to decompose the input into its wavelet coefficients. This is
    the isotropic undecimated wavelet transform implemented for multiple CPU cores. NOTE: Python is not well suited
    to multiprocessing - this may not improve execution speed.

    INPUTS:
    in1                 (no default):   Array on which the decomposition is to be performed.
    scale_count         (no default):   Maximum scale to be considered.
    scale_adjust        (default=0):    Adjustment to scale value if first scales are of no interest.
    store_smoothed      (default=False):Boolean specifier for whether the smoothed image is stored or not.
    core_count          (no default):   Indicates the number of cores to be used.

    OUTPUTS:
    detail_coeffs                       Array containing the detail coefficients.
    C0                  (optional):     Array containing the smoothest version of the input.
    """

    wavelet_filter = (1./16)*np.array([1,4,6,4,1])      # Filter-bank for use in the a trous algorithm.

    C0 = in1                                            # Sets the initial value to be the input array.

    # Initialises a zero array to store the coefficients.

    detail_coeffs = np.empty([scale_count-scale_adjust, in1.shape[0], in1.shape[1]])

    # The following loop, which iterates up to scale_adjust, applies the a trous algorithm to the scales which are
    # considered insignificant. This is important as each set of wavelet coefficients depends on the last smoothed
    # version of the input.

    if scale_adjust>0:
        for i in range(0, scale_adjust):
            C0 = mp_a_trous(C0, wavelet_filter, i, core_count)

    # The meat of the algorithm - two sequential applications fo the a trous followed by determination and storing of
    # the detail coefficients. C0 is reassigned the value of C on each loop - C0 is always the smoothest version of the
    # input image.

    for i in range(scale_adjust,scale_count):
        C = mp_a_trous(C0, wavelet_filter, i, core_count)                   # Approximation coefficients.
        C1 = mp_a_trous(C, wavelet_filter, i, core_count)                   # Approximation coefficients.
        detail_coeffs[i-scale_adjust,:,:] = C0 - C1                         # Detail coefficients.
        C0 = C

    if store_smoothed:
        return detail_coeffs, C0
    else:
        return detail_coeffs

def mp_iuwt_recomposition(in1, scale_adjust, core_count, smoothed_array):
    """
    This function calls the a trous algorithm code to recompose the input into a single array. This is the
    implementation of the isotropic undecimated wavelet transform recomposition for multiple CPU cores.

    INPUTS:
    in1             (no default):   Array containing wavelet coefficients.
    scale_adjust    (no default):   Indicates the number of omitted array pages.
    core_count      (no default):   Indicates the number of cores to be used.
    smoothed_array  (default=None): For a complete inverse transform, this must be the smoothest approximation.

    OUTPUTS:
    recomposiiton                   Array containing the reconstructed image.
    """

    wavelet_filter = (1./16)*np.array([1,4,6,4,1])      # Filter-bank for use in the a trous algorithm.

    # Determines scale with adjustment and creates a zero array to store the output, unless smoothed_array is given.

    max_scale = in1.shape[0] + scale_adjust

    if smoothed_array is None:
        recomposition = np.zeros([in1.shape[1], in1.shape[2]])
    else:
        recomposition = smoothed_array

    # The following loops call the a trous algorithm code to recompose the input. The first loop assumes that there are
    # non-zero wavelet coefficients at scales above scale_adjust, while the second loop completes the recomposition
    # on the scales less than scale_adjust.

    for i in range(max_scale-1, scale_adjust-1, -1):
        recomposition = mp_a_trous(recomposition, wavelet_filter, i, core_count) + in1[i-scale_adjust,:,:]

    if scale_adjust>0:
        for i in range(scale_adjust-1, -1, -1):
            recomposition = mp_a_trous(recomposition, wavelet_filter, i, core_count)

    return recomposition

def mp_a_trous(C0, wavelet_filter, scale, core_count):
    """
    This is a reimplementation of the a trous filter which makes use of multiprocessing. In particular,
    it divides the input array of dimensions NxN into M smaller arrays of dimensions (N/M)xN, where M is the
    number of cores which are to be used.

    INPUTS:
    C0              (no default):   The current array which is to be decomposed.
    wavelet_filter  (no default):   The filter-bank which is applied to the components of the transform.
    scale           (no default):   The scale at which decomposition is to be carried out.
    core_count      (no default):   The number of CPU cores over which the task should be divided.

    OUTPUTS:
    shared_array                    The decomposed array.
    """

    # Creates an array which may be accessed by multiple processes.

    shared_array_base = mp.Array(ctypes.c_float, C0.shape[0]**2, lock=False)
    shared_array = np.frombuffer(shared_array_base, dtype=ctypes.c_float)
    shared_array = shared_array.reshape(C0.shape)
    shared_array[:,:] = C0

    # Division of the problem and allocation of processes to cores.

    processes = []

    for i in range(core_count):
        process = mp.Process(target = mp_a_trous_kernel, args = (shared_array, wavelet_filter, scale, i,
                                                     C0.shape[0]//core_count, 'row',))
        process.start()
        processes.append(process)

    for i in processes:
        i.join()

    processes = []

    for i in range(core_count):
        process = mp.Process(target = mp_a_trous_kernel, args = (shared_array, wavelet_filter, scale, i,
                                                     C0.shape[1]//core_count, 'col',))
        process.start()
        processes.append(process)

    for i in processes:
        i.join()

    return shared_array

def mp_a_trous_kernel(C0, wavelet_filter, scale, slice_ind, slice_width, r_or_c="row"):
    """
    This is the convolution step of the a trous algorithm.

    INPUTS:
    C0              (no default):       The current array which is to be decomposed.
    wavelet_filter  (no default):       The filter-bank which is applied to the elements of the transform.
    scale           (no default):       The scale at which decomposition is to be carried out.
    slice_ind       (no default):       The index of the particular slice in which the decomposition is being performed.
    slice_width     (no default):       The number of elements of the shorter dimension of the array for decomposition.
    r_or_c          (default = "row"):  Indicates whether strips are rows or columns.

    OUTPUTS:
    NONE - C0 is a mutable array which this function alters. No value is returned.

    """

    lower_bound = slice_ind*slice_width
    upper_bound = (slice_ind+1)*slice_width

    if r_or_c == "row":
        row_conv = wavelet_filter[2]*C0[:,lower_bound:upper_bound]

        row_conv[(2**(scale+1)):,:] += wavelet_filter[0]*C0[:-(2**(scale+1)),lower_bound:upper_bound]
        row_conv[:(2**(scale+1)),:] += wavelet_filter[0]*C0[(2**(scale+1))-1::-1,lower_bound:upper_bound]

        row_conv[(2**scale):,:] += wavelet_filter[1]*C0[:-(2**scale),lower_bound:upper_bound]
        row_conv[:(2**scale),:] += wavelet_filter[1]*C0[(2**scale)-1::-1,lower_bound:upper_bound]

        row_conv[:-(2**scale),:] += wavelet_filter[3]*C0[(2**scale):,lower_bound:upper_bound]
        row_conv[-(2**scale):,:] += wavelet_filter[3]*C0[:-(2**scale)-1:-1,lower_bound:upper_bound]

        row_conv[:-(2**(scale+1)),:] += wavelet_filter[4]*C0[(2**(scale+1)):,lower_bound:upper_bound]
        row_conv[-(2**(scale+1)):,:] += wavelet_filter[4]*C0[:-(2**(scale+1))-1:-1,lower_bound:upper_bound]

        C0[:,lower_bound:upper_bound] = row_conv

    elif r_or_c == "col":
        col_conv = wavelet_filter[2]*C0[lower_bound:upper_bound,:]

        col_conv[:,(2**(scale+1)):] += wavelet_filter[0]*C0[lower_bound:upper_bound,:-(2**(scale+1))]
        col_conv[:,:(2**(scale+1))] += wavelet_filter[0]*C0[lower_bound:upper_bound,(2**(scale+1))-1::-1]

        col_conv[:,(2**scale):] += wavelet_filter[1]*C0[lower_bound:upper_bound,:-(2**scale)]
        col_conv[:,:(2**scale)] += wavelet_filter[1]*C0[lower_bound:upper_bound,(2**scale)-1::-1]

        col_conv[:,:-(2**scale)] += wavelet_filter[3]*C0[lower_bound:upper_bound,(2**scale):]
        col_conv[:,-(2**scale):] += wavelet_filter[3]*C0[lower_bound:upper_bound,:-(2**scale)-1:-1]

        col_conv[:,:-(2**(scale+1))] += wavelet_filter[4]*C0[lower_bound:upper_bound,(2**(scale+1)):]
        col_conv[:,-(2**(scale+1)):] += wavelet_filter[4]*C0[lower_bound:upper_bound,:-(2**(scale+1))-1:-1]

        C0[lower_bound:upper_bound,:] = col_conv




#! /usr/bin/env python

"""
Module for functions that check available memory, and input sizes.
"""
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['check_enough_memory',
           'get_available_memory']

from psutil import virtual_memory


def get_available_memory(verbose=True):
    """ Returns the available memory in bytes.
    """
    mem = virtual_memory()
    if verbose: 
        print("System total memory = {:.3f} GB".format(mem.total/1e9))
        print("System available memory = {:.3f} GB".format(mem.available/1e9))
    return mem.available


def check_enough_memory(input_bytes, factor=1, verbose=True):
    """ Checks if the system's available memory is larger than 
    factor*input_bytes. This function is used to check the inputs of algorithms
    and avoid system/Python crashes or heavy swapping.
    
    Parameters
    ----------
    input_bytes : float
        The size in bytes of the inputs of a given function.
    factor : float
        Scales how much memory is needed in terms of the size of input_bytes.
    """
    mem = virtual_memory()
    load = factor*input_bytes
    if load>mem.total:
        return False        # total physical memory is smaller than input
    if verbose:  
        print("System available memory = {:} bytes".format(mem.available))
    if load>=mem.available:
        return False        # available memory is not enough
    else:
        return True         # enough available memory
    
    


"""
In subpackage ``pca`` are the PCA based algorithms (Soumer et al. 2012 and Amara
et al. 2012) for cubes reference-PSF/background subtraction. PCA comes in
different flavors and with speed modifications:

- *Full-frame PCA*, using the whole cube as the PCA reference library in the
case of ADI or SDI (ISF cube), or a sequence of reference frames (reference
star) in the case of RDI. For ADI a big data matrix NxP, where N is the number
of frames and P the number of pixels in a frame is created. Then PCA is done
through eigen-decomposition of the covariance matrix (~$DD^T$) or the SVD of
the centered data matrix. SVD can be calculated using different libraries
including a very fast one (default option): randomized SVD (Halko et al. 2009).

- *Annular PCA*, and *subannular PCA* (quadrants of annulus) perform a local PCA
taking into account a parallactic angle rejection for allowing FOV rotation and
avoid planet self-subtraction. These local PCA algorithms process many (number
of patches times number of frames) smaller matrices increasing the computation
time.

- *parallel subannular PCA* which uses multiprocessing and allows to distribute
the computations amongst the available cores in your machine.
"""
from __future__ import absolute_import

from .pca_fullfr import *
from .pca_local import *
from .svd import *
from .utils_pca import *



#! /usr/bin/env python

"""
Module with functions for posterior sampling of the NEGFC parameters using
nested sampling (``nestle``).
"""

from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez',
__all__ = ['nested_negfc_sampling',
           'nested_sampling_results']

import nestle
import corner
import numpy as np
from matplotlib import pyplot as plt
from ..conf import time_ini, timing
from .mcmc_sampling import lnlike, confidence, show_walk_plot


def nested_negfc_sampling(init, cube, angs, plsc, psf, fwhm, annulus_width=2,
                          aperture_radius=1, ncomp=10, scaling=None,
                          svd_mode='lapack', cube_ref=None, collapse='median',
                          w=(5, 5, 200), method='single', npoints=100,
                          dlogz=0.1, decline_factor=None, rstate=None,
                          verbose=True):
    """ Runs a nested sampling algorithm in order to determine the position and
    the flux of the planet using the 'Negative Fake Companion' technique. The
    result of this procedure is a a ``nestle`` object containing the samples
    from the posterior distributions of each of the 3 parameters. It provides
    pretty good results (value plus error bars) compared to a more CPU intensive
    Monte Carlo approach with the affine invariant sampler (``emcee``).

    Parameters
    ----------
    init: array_like or tuple of length 3
        The first guess for the position and flux of the planet, respectively.
        It serves for generating the bounds of the log prior function (uniform
        in a bounded interval).
    cube: array_like
        Frame sequence of cube.
    angs: array_like
        The relative path to the parallactic angle fits image or the angs itself.
    plsc: float
        The platescale, in arcsec per pixel.
    psf: array_like
        The PSF template. It must be centered and the flux in a 1*FWHM aperture
        must equal 1.
    fwhm : float
        The FHWM in pixels.
    annulus_width: float, optional
        The width in pixel of the annulus on which the PCA is performed.
    aperture_radius: float, optional
        The radius of the circular aperture.
    ncomp: int optional
        The number of principal components.
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With
        "temp-mean" then temporal px-wise mean subtraction is done and with
        "temp-standard" temporal mean centering plus scaling to unit variance
        is done.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    cube_ref: array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    w : tuple of length 3
        The size of the bounds (around the initial state ``init``) for each
        parameter.
    method : {"single", "multi", "classic"}, str optional
        Flavor of nested sampling. Single ellipsoid works well for the NEGFC and
        is the default.
    npoints : int optional
        Number of active points. At least ndim+1 (4 will produce bad results).
        For problems with just a few parameters (<=5) like the NEGFC, good
        results are obtained with 100 points (default).
    dlogz : Estimated remaining evidence
        Iterations will stop when the estimated contribution of the remaining
        prior volume to the total evidence falls below this threshold.
        Explicitly, the stopping criterion is log(z + z_est) - log(z) < dlogz
        where z is the current evidence from all saved samples, and z_est is the
        estimated contribution from the remaining volume. This option and
        decline_factor are mutually exclusive. If neither is specified, the
        default is dlogz=0.5.
    decline_factor : float, optional
        If supplied, iteration will stop when the weight (likelihood times prior
        volume) of newly saved samples has been declining for
        decline_factor * nsamples consecutive samples. A value of 1.0 seems to
        work pretty well.
    rstate : random instance, optional
        RandomState instance. If not given, the global random state of the
        numpy.random module will be used.

    Returns
    -------
    res : nestle object
        ``Nestle`` object with the nested sampling results, including the
        posterior samples.

    Notes
    -----
    Nested Sampling is a computational approach for integrating posterior
    probability in order to compare models in Bayesian statistics. It is similar
    to Markov Chain Monte Carlo (MCMC) in that it generates samples that can be
    used to estimate the posterior probability distribution. Unlike MCMC, the
    nature of the sampling also allows one to calculate the integral of the
    distribution. It also happens to be a pretty good method for robustly
    finding global maxima.

    Nestle documentation:
    http://kbarbary.github.io/nestle/

    Convergence:
    http://kbarbary.github.io/nestle/stopping.html
    Nested sampling has no well-defined stopping point. As iterations continue,
    the active points sample a smaller and smaller region of prior space.
    This can continue indefinitely. Unlike typical MCMC methods, we don't gain
    any additional precision on the results by letting the algorithm run longer;
    the precision is determined at the outset by the number of active points.
    So, we want to stop iterations as soon as we think the active points are
    doing a pretty good job sampling the remaining prior volume - once we've
    converged to the highest-likelihood regions such that the likelihood is
    relatively flat within the remaining prior volume.

    Method:
    The trick in nested sampling is to, at each step in the algorithm,
    efficiently choose a new point in parameter space drawn with uniform
    probability from the parameter space with likelihood greater than the
    current likelihood constraint. The different methods all use the
    current set of active points as an indicator of where the target
    parameter space lies, but differ in how they select new points from  it.
    "classic" is close to the method described in Skilling (2004).
    "single", Mukherjee, Parkinson & Liddle (2006), Determines a single
    ellipsoid that bounds all active points,
    enlarges the ellipsoid by a user-settable factor, and selects a new point
    at random from within the ellipsoid.
    "multiple", Shaw, Bridges & Hobson (2007) and Feroz, Hobson & Bridges 2009
    (Multinest). In cases where the posterior is multi-modal,
    the single-ellipsoid method can be extremely inefficient: In such
    situations, there are clusters of active points on separate
    high-likelihood regions separated by regions of lower likelihood.
    Bounding all points in a single ellipsoid means that the ellipsoid
    includes the lower-likelihood regions we wish to avoid
    sampling from.
    The solution is to detect these clusters and bound them in separate
    ellipsoids. For this, we use a recursive process where we perform
    K-means clustering with K=2. If the resulting two ellipsoids have a
    significantly lower total volume than the parent ellipsoid (less than half),
    we accept the split and repeat the clustering and volume test on each of
    the two subset of points. This process continues recursively.
    Alternatively, if the total ellipse volume is significantly greater
    than expected (based on the expected density of points) this indicates
    that there may be more than two clusters and that K=2 was not an
    appropriate cluster division.
    We therefore still try to subdivide the clusters recursively. However,
    we still only accept the final split into N clusters if the total volume
    decrease is significant.

    """

    def prior_transform(x):
        """ x:[0,1]

        The prior transform is dinamically created with these bound:
        [radius-w1:radius+w1], [theta-w2:theta+w2], [flux-w3:flux+w3]

        Notes
        -----
        The prior transform function is used to specify the Bayesian prior for the
        problem, in a round-about way. It is a transformation from a space where
        variables are independently and uniformly distributed between 0 and 1 to
        the parameter space of interest. For independent parameters, this would be
        the product of the inverse cumulative distribution function (also known as
        the percent point function or quantile function) for each parameter.
        http://kbarbary.github.io/nestle/prior.html

        """
        a1 = 2 * w[0]
        a2 = init[0] - w[0]
        b1 = 2 * w[1]
        b2 = init[1] - w[1]
        c1 = 2 * w[2]
        c2 = init[2] - w[2]
        return np.array([a1 * x[0] + a2, b1 * x[1] + b2, c1 * x[2] + c2])

    def f(param):
        return lnlike(param=param, cube=cube, angs=angs, plsc=plsc,
                      psf_norm=psf, fwhm=fwhm, annulus_width=annulus_width,
                      aperture_radius=aperture_radius, initial_state=init,
                      cube_ref=cube_ref, svd_mode=svd_mode, scaling=scaling,
                      fmerit='sum', ncomp=ncomp, collapse=collapse)

    # -------------------------------------------------------------------------
    if verbose:  start = time_ini()

    if verbose:
        print('Prior bounds on parameters:')
        print('Radius [{},{}]'.format(init[0] - w[0], init[0] + w[0], ))
        print('Theta [{},{}]'.format(init[1] - w[1], init[1] + w[1]))
        print('Flux [{},{}]'.format(init[2] - w[2], init[2] + w[2]))
        print('\nUsing {:} active points'.format(npoints))

    res = nestle.sample(f, prior_transform, ndim=3, method=method,
                        npoints=npoints, rstate=rstate, dlogz=dlogz,
                        decline_factor=decline_factor)

    # if verbose:  print; timing(start)
    if verbose:
        print('\nTotal running time:')
        timing(start)
    return res


def nested_sampling_results(ns_object, burnin=0.4, bins=None):
    """ Shows the results of the Nested Sampling, summary, parameters with errors,
    walk and corner plots.
    """
    res = ns_object
    nsamples = res.samples.shape[0]
    indburnin = np.percentile(np.array(range(nsamples)), burnin * 100)

    print(res.summary())

    print(
        '\nNatural log of prior volume and Weight corresponding to each sample')
    plt.figure(figsize=(12, 4))
    plt.subplot(1, 2, 1)
    plt.plot(res.logvol, '.', alpha=0.5, color='gray')
    plt.xlabel('samples')
    plt.ylabel('logvol')
    plt.vlines(indburnin, res.logvol.min(), res.logvol.max(),
               linestyles='dotted')
    plt.subplot(1, 2, 2)
    plt.plot(res.weights, '.', alpha=0.5, color='gray')
    plt.xlabel('samples')
    plt.ylabel('weights')
    plt.vlines(indburnin, res.weights.min(), res.weights.max(),
               linestyles='dotted')
    plt.show()

    print("\nWalk plots before the burnin")
    show_walk_plot(np.expand_dims(res.samples, axis=0))
    if burnin > 0:
        print("\nWalk plots after the burnin")
        show_walk_plot(np.expand_dims(res.samples[indburnin:], axis=0))

    mean, cov = nestle.mean_and_cov(res.samples[indburnin:],
                                    res.weights[indburnin:])
    print("\nWeighted mean +- sqrt(covariance)")
    print("Radius = {:.3f} +/- {:.3f}".format(mean[0], np.sqrt(cov[0, 0])))
    print("Theta = {:.3f} +/- {:.3f}".format(mean[1], np.sqrt(cov[1, 1])))
    print("Flux = {:.3f} +/- {:.3f}".format(mean[2], np.sqrt(cov[2, 2])))

    if bins is None:
        bins = int(np.sqrt(res.samples[indburnin:].shape[0]))
        print("\nHist bins =", bins)
    ranges = None

    fig = corner.corner(res.samples[indburnin:], bins=bins,
                        labels=["$r$", r"$\theta$", "$f$"],
                        weights=res.weights[indburnin:], range=ranges,
                        plot_contours=True)
    fig.set_size_inches(8., 8.)

    print('\nConfidence intervals')
    _ = confidence(res.samples[indburnin:], cfd=68, bins=bins,
                   weights=res.weights[indburnin:],
                   gaussianFit=True, verbose=True, save=False)


#! /usr/bin/env python

"""
Module with a frame differencing algorithm for ADI post-processing.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['xloci']

import numpy as np
import scipy as sp
import pandas as pn
import itertools as itt
from multiprocessing import Pool, cpu_count
from sklearn.metrics import pairwise_distances
from ..var import get_annulus_segments
from ..preproc import cube_derotate, cube_collapse, check_pa_vector
from ..conf import time_ini, timing
from ..madi.adi_utils import _find_indices, _define_annuli
from ..conf.utils_conf import eval_func_tuple as EFT


array = None


def xloci(cube, angle_list, fwhm=4, metric='manhattan', dist_threshold=50,
          delta_rot=0.5, radius_int=0, asize=4, n_segments=4, nproc=1,
          solver='lstsq', tol=1e-3, verbose=True, full_output=False):
    """ LOCI style algorithm that models a PSF (for ADI data) with a
    least-square combination of neighbouring frames (solving the equation
    a x = b by computing a vector x of coefficients that minimizes the
    Euclidean 2-norm || b - a x ||^2).
    
    Parameters
    ----------
    cube : array_like, 3d
        Input cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.
    fwhm : float, optional
        Known size of the FHWM in pixels to be used. Default is 4.
    metric : str, optional
        Distance metric to be used ('cityblock', 'cosine', 'euclidean', 'l1',
        'l2', 'manhattan', 'correlation', etc). It uses the scikit-learn
        function ``sklearn.metrics.pairwise.pairwise_distances`` (check its
        documentation).
    dist_threshold : int
        Indices with a distance larger thatn ``dist_threshold`` percentile will
        initially discarded.
    delta_rot : int
        Minimum parallactic angle distance between the pairs.
    radius_int : int, optional
        The radius of the innermost annulus. By default is 0, if >0 then the
        central circular area is discarded.
    asize : int, optional
        The size of the annuli, in pixels.
    n_segments : int or list of ints, optional
        The number of segments for each annulus. When a single integer is given
        it is used for all annuli.
    nproc : None or int, optional
        Number of processes for parallel computing. If None the number of
        processes will be set to (cpu_count()/2). By default the algorithm works
        in single-process mode.
    solver : {'lstsq', 'nnls'}, str optional
        Choosing the solver of the least squares problem. ``lstsq`` uses the
        standard scipy least squares solver. ``nnls`` uses the scipy
        non-negative least squares solver.
    tol : float, optional
        Valid when ``solver`` is set to lstsq. Sets the cutoff for 'small'
        singular values; used to determine effective rank of a. Singular values
        smaller than ``tol * largest_singular_value`` are considered zero.
        Smaller values of ``tol`` lead to smaller residuals (more aggressive
        subtraction).
    verbose: bool, optional
        If True prints info to stdout.
    full_output: bool, optional
        Whether to return the final median combined image only or with other
        intermediate arrays.
        
    Returns
    -------
    frame_der_median : array_like, 2d
        Median combination of the de-rotated cube of residuals.

    If ``full_output`` is True, the following intermediate arrays are returned:
    cube_res, cube_der, frame_der_median

    """
    global array
    array = cube

    if verbose:
        start_time = time_ini()

    y = array.shape[1]
    if not asize < np.floor((y / 2)):
        raise ValueError("asize is too large")

    angle_list = check_pa_vector(angle_list)
    n_annuli = int(np.floor((y / 2 - radius_int) / asize))
    if verbose:
        msg = "{:} annuli. Performing least-square combination and "
        msg += "subtraction:\n"
        print(msg.format(n_annuli))

    if nproc is None:   # Hyper-threading "duplicates" the cores -> cpu_count/2
        nproc = (cpu_count()/2)

    annulus_width = asize
    if isinstance(n_segments, int):
        n_segments = [n_segments for _ in range(n_annuli)]
    elif n_segments == 'auto':
        n_segments = []
        n_segments.append(2)    # for first annulus
        n_segments.append(3)    # for second annulus
        ld = 2 * np.tan(360/4/2) * annulus_width
        for i in range(2, n_annuli):    # rest of annuli
            radius = i * annulus_width
            ang = np.rad2deg(2 * np.arctan(ld / (2 * radius)))
            n_segments.append(int(np.ceil(360/ang)))

    # annulus-wise least-squares combination and subtraction
    cube_res = np.zeros((array.shape[0], array.shape[1], array.shape[2]))
    for ann in range(n_annuli):
        n_segments_ann = n_segments[ann]
        inner_radius_ann = radius_int + ann*annulus_width

        indices = get_annulus_segments(array[0], inner_radius=inner_radius_ann,
                                       width=asize, nsegm=n_segments_ann)

        res_ann = _leastsq_ann(indices, ann, n_annuli, fwhm, angle_list,
                               delta_rot, metric, dist_threshold, radius_int,
                               asize, n_segments_ann, nproc, solver, tol,
                               verbose)

        for j in range(n_segments_ann):
            yy = indices[j][0]
            xx = indices[j][1]
            cube_res[:, yy, xx] = res_ann[j]

    cube_der = cube_derotate(cube_res, angle_list)
    frame_der_median = cube_collapse(cube_der, 'median')

    if verbose:
        print('Done processing annuli')
        timing(start_time)

    if full_output:
        return cube_res, cube_der, frame_der_median
    else:
        return frame_der_median


def _leastsq_ann(indices, ann, n_annuli, fwhm, angles, delta_rot, metric,
                 dist_threshold, radius_int, asize, n_segments_ann, nproc,
                 solver, tol, verbose):
    """ Helper function for xloci. Least-squares combination and subtraction
    for each segment in an annulus, applying a rotation threshold.
    """
    start_time = time_ini(False)

    pa_threshold, _, _ = _define_annuli(angles, ann, n_annuli, fwhm, radius_int,
                                        asize, delta_rot, n_segments_ann,
                                        verbose)
    res = []
    if nproc == 1:
        for j in range(n_segments_ann):
            segm_res = _leastsq_patch(j, indices, angles, pa_threshold, metric,
                                      dist_threshold, solver, tol)
            res.append(segm_res)

    elif nproc > 1:
        pool = Pool(processes=int(nproc))
        res = pool.map(EFT, zip(itt.repeat(_leastsq_patch),
                                range(n_segments_ann), itt.repeat(indices),
                                itt.repeat(angles), itt.repeat(pa_threshold),
                                itt.repeat(metric), itt.repeat(dist_threshold),
                                itt.repeat(solver), itt.repeat(tol)))
        pool.close()

    else:
        raise ValueError("nproc must be a positive integer")

    if verbose and nproc == 1:
        timing(start_time)

    return res


def _leastsq_patch(nseg, indices, angles, pa_threshold, metric, dist_threshold,
                   solver, tol):
    """ Helper function for _leastsq_ann.
    """
    yy = indices[nseg][0]
    xx = indices[nseg][1]
    values = array[:, yy, xx]  # n_frames x n_pxs_segment
    n_frames = array.shape[0]

    mat_dists_ann_full = pairwise_distances(values, metric=metric)

    if pa_threshold > 0:
        mat_dists_ann = np.zeros_like(mat_dists_ann_full)
        for i in range(n_frames):
            ind_fr_i = _find_indices(angles, i, pa_threshold, None, False)
            mat_dists_ann[i][ind_fr_i] = mat_dists_ann_full[i][ind_fr_i]
    else:
        mat_dists_ann = mat_dists_ann_full

    threshold = np.percentile(mat_dists_ann[mat_dists_ann != 0.0],
                              dist_threshold)
    mat_dists_ann[mat_dists_ann > threshold] = np.nan
    mat_dists_ann[mat_dists_ann == 0.0] = np.nan

    matrix_res = np.zeros((values.shape[0], yy.shape[0]))
    for i in range(n_frames):
        vector = pn.DataFrame(mat_dists_ann[i])
        vector.columns = ['i']
        if vector.sum().values != 0:
            ind_ref = np.where(~np.isnan(vector))[0]
            A = values[ind_ref]
            b = values[i]
            if solver == 'lstsq':
                coef = sp.linalg.lstsq(A.T, b, cond=tol)[0]     # SVD method
            elif solver == 'nnls':
                coef = sp.optimize.nnls(A.T, b)[0]
            elif solver == 'lsq':
                coef = sp.optimize.lsq_linear(A.T, b, bounds=(0, 1),
                                              method='trf',
                                              lsq_solver='lsmr')['x']
            else:
                raise ValueError("solver not recognized")

        recon = np.dot(coef, values[ind_ref])
        matrix_res[i] = values[i] - recon

    return matrix_res




#! /usr/bin/env python

"""
Module with a frame differencing algorithm for ADI post-processing.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['frame_diff']

import numpy as np
import pandas as pn
import itertools as itt
from multiprocessing import Pool, cpu_count
from sklearn.metrics import pairwise_distances
from ..var import get_annulus_segments, pp_subplots
from ..preproc import cube_derotate, cube_collapse, check_pa_vector
from ..conf import time_ini, timing
from ..pca.utils_pca import pca_annulus
from ..madi.adi_utils import _compute_pa_thresh, _find_indices, _define_annuli
from ..conf.utils_conf import eval_func_tuple as EFT


array = None


def frame_diff(cube, angle_list, fwhm=4, metric='manhattan', dist_threshold=50,
               n_similar=None, delta_rot=0.5, radius_int=2, asize=4, ncomp=None,
               nproc=1, verbose=True, debug=False):
    """ Frame differencing algorithm. It uses vector distance (depending on
    ``metric``), using separately the pixels from different annuli of ``asize``
    width, to create pairs of most similar images. Then it performs pair-wise
    subtraction and combines the residuals.
    
    Parameters
    ----------
    cube : array_like, 3d
        Input cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.
    fwhm : float, optional
        Known size of the FHWM in pixels to be used. Default is 4.
    metric : str, optional
        Distance metric to be used ('cityblock', 'cosine', 'euclidean', 'l1',
        'l2', 'manhattan', 'correlation', etc). It uses the scikit-learn
        function ``sklearn.metrics.pairwise.pairwise_distances`` (check its
        documentation).
    dist_threshold : int
        Indices with a distance larger thatn ``dist_threshold`` percentile will
        initially discarded.
    n_similar : None or int, optional
        If a postive integer value is given, then a median combination of
        ``n_similar`` frames will be used instead of the most similar one.
    delta_rot : int
        Minimum parallactic angle distance between the pairs.
    radius_int : int, optional
        The radius of the innermost annulus. By default is 0, if >0 then the
        central circular area is discarded.
    asize : int, optional
        The size of the annuli, in pixels.
    ncomp : None or int, optional
        If a positive integer value is given, then the annulus-wise PCA low-rank
        approximation with ``ncomp`` principal components will be subtracted.
        The pairwise subtraction will be performed on these residuals.
    nproc : None or int, optional
        Number of processes for parallel computing. If None the number of
        processes will be set to (cpu_count()/2). By default the algorithm works
        in single-process mode.
    verbose: bool, optional
        If True prints info to stdout.
    debug : bool, optional
        If True the distance matrices will be plotted and additional information
        will be given.
        
    Returns
    -------
    final_frame : array_like, 2d
        Median combination of the de-rotated cube.
    """
    global array
    array = cube

    if verbose:
        start_time = time_ini()

    y = array.shape[1]
    if not asize < np.floor((y / 2)):
        raise ValueError("asize is too large")

    angle_list = check_pa_vector(angle_list)
    n_annuli = int(np.floor((y / 2 - radius_int) / asize))
    if verbose:
        if ncomp is not None:
            msg = "{:} annuli. Performing annular PCA subtraction with {:} PCs "
            msg += "and pair-wise subtraction:\n"
            print(msg.format(n_annuli, ncomp))
        else:
            msg = "{:} annuli. Performing pair-wise subtraction:\n"
            print(msg.format(n_annuli))

    if nproc is None:   # Hyper-threading "duplicates" the cores -> cpu_count/2
        nproc = (cpu_count()/2)

    # annulus-wise pair-wise subtraction
    final_frame = []
    if nproc == 1:
        for ann in range(n_annuli):
            res_ann = _pairwise_ann(ann, n_annuli, fwhm, angle_list, delta_rot,
                                    metric, dist_threshold, n_similar,
                                    radius_int, asize, ncomp, verbose, debug)
            final_frame.append(res_ann)
    elif nproc > 1:
        pool = Pool(processes=int(nproc))
        res = pool.map(EFT, zip(itt.repeat(_pairwise_ann), range(n_annuli),
                                itt.repeat(n_annuli), itt.repeat(fwhm),
                                itt.repeat(angle_list), itt.repeat(delta_rot),
                                itt.repeat(metric), itt.repeat(dist_threshold),
                                itt.repeat(n_similar), itt.repeat(radius_int),
                                itt.repeat(asize), itt.repeat(ncomp),
                                itt.repeat(verbose), itt.repeat(debug),
                                itt.repeat(nproc)))
        final_frame = np.array(res)
        pool.close()
    final_frame = np.sum(final_frame, axis=0)

    if verbose:
        print('Done processing annuli')
        timing(start_time)

    return final_frame


def _pairwise_ann(ann, n_annuli, fwhm, angles, delta_rot, metric,
                  dist_threshold, n_similar, radius_int, asize, ncomp, verbose,
                  debug=False, nproc=1):
    """
    Helper functions for pair-wise subtraction for a single annulus.
    """
    start_time = time_ini(False)

    n_frames = array.shape[0]

    pa_threshold, in_rad, ann_center = _define_annuli(angles, ann, n_annuli,
                                                      fwhm, radius_int, asize,
                                                      delta_rot, 1, verbose)
    if ncomp is not None:
        arrayin = pca_annulus(array, None, ncomp, asize, ann_center,
                              svd_mode='lapack', scaling=None, collapse=None)
    else:
        arrayin = array

    yy, xx = get_annulus_segments(array[0], inner_radius=in_rad, width=asize,
                                  nsegm=1)[0]
    values = arrayin[:, yy, xx]  # n_frames x n_pxs_annulus

    if debug:
        print('Done taking pixel intensities from annulus.')
        timing(start_time)

    mat_dists_ann_full = pairwise_distances(values, metric=metric)

    if pa_threshold > 0:
        mat_dists_ann = np.zeros_like(mat_dists_ann_full)
        for i in range(n_frames):
            ind_fr_i = _find_indices(angles, i, pa_threshold, None, False)
            mat_dists_ann[i][ind_fr_i] = mat_dists_ann_full[i][ind_fr_i]
    else:
        mat_dists_ann = mat_dists_ann_full

    if debug:
        #pp_subplots(mat_dists_ann)
        msg = 'Done calculating the {:} distance for annulus {:}'
        print(msg.format(metric, ann+1))
        timing(start_time)

    threshold = np.percentile(mat_dists_ann[mat_dists_ann != 0.0],
                              dist_threshold)
    mat_dists_ann[mat_dists_ann > threshold] = np.nan
    mat_dists_ann[mat_dists_ann == 0.0] = np.nan
    if not mat_dists_ann[~np.isnan(mat_dists_ann)].size > 0:
        raise RuntimeError('No pairs left. Decrease thresholds')

    if debug:
        pp_subplots(mat_dists_ann)
        print('Done thresholding/checking distances.')
        timing(start_time)

    # median of n ``n_similar`` most similar patches
    cube_res = []
    if n_similar is not None:
        if not n_similar >= 3:
            raise ValueError("n_similar must be >= 3 or None")
        for i in range(n_frames):
            vector = pn.DataFrame(mat_dists_ann[i])
            vector.columns = ['i']
            if vector.sum().values == 0:
                continue
            else:
                vector_sorted = vector.i.sort_values()[:n_similar]
                ind_n_similar = vector_sorted.index.values
                # median subtraction
                res = values[i] - np.median((values[ind_n_similar]), axis=0)
                cube_res.append(res)

        cube_res = np.array(cube_res)
        angles_list = angles

    # taking just the most similar frame
    else:
        ind = []
        for i in range(n_frames):
            vector = pn.DataFrame(mat_dists_ann[i])
            vector.columns = ['i']
            if vector.sum().values == 0:
                continue
            else:
                ind.append((i, vector.idxmin().tolist()[0]))
                ind.append((vector.idxmin().tolist()[0], i))

        if debug:
            print('Done finding pairs. Total found: ', len(ind)/2)
            timing(start_time)

        df = pn.DataFrame(ind)  # sorting using pandas dataframe
        df.columns = ['i', 'j']
        df = df.sort_values('i')

        indices = df.values
        indices = indices.astype(int)  # back to a ndarray int type

        size = indices.shape[0]
        angles_list = np.zeros((size))
        for i in range(size):
            angles_list[i] = angles[indices[i][0]]  # filter of the angles vector

        cube_res = np.zeros((size, yy.shape[0]))
        # pair-wise subtraction
        for i in range(size):
            res = values[indices[i][0]] - values[indices[i][1]]
            cube_res[i] = res

    cube_out = np.zeros((cube_res.shape[0], array.shape[1], array.shape[2]))
    cube_out[:, yy, xx] = cube_res
    cube_der = cube_derotate(cube_out, angles_list)
    frame_der_median = cube_collapse(cube_der, 'median')

    if verbose and nproc == 1:
        timing(start_time)

    return frame_der_median


def _pw_rot_res(cube, angle_list, fwhm=4, delta_rot=0.5, inner_radius=2,
                asize=4, verbose=True, debug=False):
    """

    Parameters
    ----------
    cube : array_like, 3d
        Input cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.
    fwhm : float, optional
        Known size of the FHWM in pixels to be used. Default is 4.
    delta_rot : int
        Minimum parallactic angle distance between the pairs.
    inner_radius : int, optional
        The radius of the innermost annulus. By default is 0, if >0 then the
        central circular area is discarded.
    asize : int, optional
        The size of the annuli, in pixels.
    verbose: bool, optional
        If True prints info to stdout.
    debug : bool, optional
        If True the distance matrices will be plotted and additional information
        will be given.

    Returns
    -------
    final_frame : array_like, 2d

    """
    array = cube

    if verbose:
        start_time = time_ini()

    n_frames = array.shape[0]
    y = array.shape[1]
    if not asize < np.floor((y / 2)):
        raise ValueError("asize is too large")

    angle_list = check_pa_vector(angle_list)

    ann_center = (inner_radius + (asize / 2.0))
    pa_threshold = _compute_pa_thresh(ann_center, fwhm, delta_rot)
    if debug:
        print(pa_threshold)

    # annulus-wise pair-wise subtraction
    res = []
    for i in range(n_frames):
        indp, indn = _find_indices(angle_list, i, pa_threshold,
                                   out_closest=True)
        if debug:
            print(indp, indn)

        res.append(array[i] - array[indn])
        if indn == n_frames-1: break

    if verbose:
        print('Done processing annulus')
        timing(start_time)

    return np.array(res)



#! /usr/bin/env python

"""
Module with the function of merit definitions for the NEGFC optimization.
"""
from __future__ import print_function

__all__ = []

import numpy as np
from skimage.draw import circle
from ..phot import cube_inject_companions
from ..var import frame_center
from ..pca.utils_pca import pca_annulus


def chisquare(modelParameters, cube, angs, plsc, psfs_norm, fwhm, annulus_width,  
              aperture_radius, initialState, ncomp, cube_ref=None, 
              svd_mode='lapack', scaling=None, fmerit='sum', collapse='median',
              imlib='opencv', interpolation='lanczos4'):
    """
    Calculate the reduced chi2:
    \chi^2_r = \frac{1}{N-3}\sum_{j=1}^{N} |I_j|,
    where N is the number of pixels within a circular aperture centered on the 
    first estimate of the planet position, and I_j the j-th pixel intensity.
    
    Parameters
    ----------    
    modelParameters: tuple
        The model parameters, typically (r, theta, flux).
    cube: numpy.array
        The cube of fits images expressed as a numpy.array.
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array. 
    plsc: float
        The platescale, in arcsec per pixel.
    psfs_norm: numpy.array
        The scaled psf expressed as a numpy.array.    
    fwhm : float
        The FHWM in pixels.
    annulus_width: int, optional
        The width in terms of the FWHM of the annulus on which the PCA is done.       
    aperture_radius: int, optional
        The radius of the circular aperture in terms of the FWHM.
    initialState: numpy.array
        Position (r, theta) of the circular aperture center.
    ncomp: int
        The number of principal components.
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.         
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Chooses the figure of merit to be used. stddev works better for close in
        companions sitting on top of speckle noise.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
        
    Returns
    -------
    out: float
        The reduced chi squared.
        
    """    
    try:
        r, theta, flux = modelParameters
    except TypeError:
        print('paraVector must be a tuple, {} was given'.format(type(modelParameters)))

    # Create the cube with the negative fake companion injected
    cube_negfc = cube_inject_companions(cube, psfs_norm, angs, flevel=-flux,
                                        plsc=plsc, rad_dists=[r], n_branches=1,
                                        theta=theta, imlib=imlib, verbose=False,
                                        interpolation=interpolation)
                                      
    # Perform PCA and extract the zone of interest
    values = get_values_optimize(cube_negfc, angs, ncomp, annulus_width*fwhm,
                                 aperture_radius*fwhm, initialState[0],
                                 initialState[1], cube_ref=cube_ref, 
                                 svd_mode=svd_mode, scaling=scaling,
                                 collapse=collapse)
    
    # Function of merit
    if fmerit=='sum':
        values = np.abs(values)
        chi2 = np.sum(values[values>0])
        N =len(values[values>0])    
        return chi2/(N-3)
    elif fmerit=='stddev':
        return np.std(values[values!=0]) 
    else:
        raise RuntimeError('fmerit choice not recognized')       



def get_values_optimize(cube, angs, ncomp, annulus_width, aperture_radius, 
                        r_guess, theta_guess, cube_ref=None, svd_mode='lapack',
                        scaling=None, imlib='opencv', interpolation='lanczos4',
                        collapse='median', debug=False):
    """
    Extracts a PCA-ed annulus from the cube and returns the flux values of the 
    pixels included in a circular aperture centered at a given position.
    
    Parameters
    ----------
    cube: numpy.array
        The cube of fits images expressed as a numpy.array.
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array.
    ncomp: int
        The number of principal component.
    annulus_width: float
        The width in pixels of the annulus on which the PCA is performed.
    aperture_radius: float
        The radius in pixels of the circular aperture.
    r_guess: float
        The radial position of the center of the circular aperture. This 
        parameter is NOT the radial position of the candidate associated to the 
        Markov chain, but should be the fixed initial guess.
    theta_guess: float
        The angular position of the center of the circular aperture. This 
        parameter is NOT the angular position of the candidate associated to the 
        Markov chain, but should be the fixed initial guess.  
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {None, 'temp-mean', 'temp-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is returned.
    debug: boolean
        If True, the cube is returned along with the values.        
        
    Returns
    -------
    values: numpy.array
        The pixel values in the circular aperture after the PCA process.

    If debug is True the PCA frame is returned (in case when collapse is not None)
        
    """
    centy_fr, centx_fr = frame_center(cube[0])
    posy = r_guess * np.sin(np.deg2rad(theta_guess)) + centy_fr
    posx = r_guess * np.cos(np.deg2rad(theta_guess)) + centx_fr
    halfw = max(aperture_radius, annulus_width/2.)

    # Checking annulus/aperture sizes. Assuming square frames
    msg = 'The annulus and/or the circular aperture used by the NegFC falls '
    msg += 'outside the FOV. Try increasing the size of your frames or '
    msg += 'decreasing the annulus or aperture size'
    if r_guess > centx_fr-halfw or r_guess <= halfw:
        raise RuntimeError(msg)
        
    pca_res = pca_annulus(cube, angs, ncomp, annulus_width, r_guess, cube_ref,
                          svd_mode, scaling, imlib=imlib,
                          interpolation=interpolation, collapse=collapse)
    indices = circle(posy, posx, radius=aperture_radius)
    yy, xx = indices

    if collapse is None:
        values = pca_res[:, yy, xx].ravel()
    else:
        values = pca_res[yy, xx].ravel()

    if debug and collapse is not None:
        return values, pca_res
    else:
        return values




#! /usr/bin/env python

"""
Module with a function for displaying numpy arrays on DS9 (wrapper of pyds9).
"""
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'

import warnings
try:
    import pyds9
    no_pyds9 = False
    __all__ = ['ds9']
except ImportError:
    msg = "Pyds9 not available."
    warnings.warn(msg, ImportWarning)
    no_pyds9 = True
    __all__ = []

class ds9(object):
    """ Creates a DS9 window (named 'VIP_ds9') using pyDS9 functionality.

    The methods of this class allow interacting with the DS9 window. It uses
    XPA under the hood, which mean that using the ``set`` and ``get`` methods
    one can have fine control of the DS9 window. More info here:
    http://ds9.si.edu/doc/ref/xpa.html.

    """
    def __init__(self):
        """ __init__ method.
        """
        self.window_name = 'VIP_ds9'
        self.window = pyds9.DS9(self.window_name)

    def clear_frames(self):
        """ Clears all frames. """
        self.window = pyds9.DS9(self.window_name)
        self.window.set('frame clear all')

    def create_frame(self):
        """ Creates a new frame (if no window it creates one). """
        self.window = pyds9.DS9(self.window_name)
        self.window.set('frame new')

    def cmap(self, value=None):
        """ Controls the colormap for the current frame. The colormap name is
        not case sensitive. A valid contrast value is  from 0 to 10 and bias
        value from 0 to 1.

        Parameters
        ----------
        value : str, optional
            Value to be passed to the cmap command. Syntax:
            cmap [<colormap>]
                 [file]
                 [load <filename>]
                 [save <filename>]
                 [invert yes|no]
                 [value <constrast> <bias>]
                 [tag [load|save] <filename>]
                 [tag delete]
                 [match]
                 [lock [yes|no]]
                 [open|close]
        """
        self.window = pyds9.DS9(self.window_name)
        self.window.set('cmap ' + str(value))


    def crosshair_get(self):
        """ Gets the physical coordinates (x,y) of the crosshair.

        Returns
        -------
        x : float, optional
            Crosshair x coordinate.
        y : float, optional
            Crosshair y coordinate.
        """
        self.window = pyds9.DS9(self.window_name)
        coor = self.window.get('crosshair')
        x = float(coor.split()[0])
        y = float(coor.split()[1])
        return x, y

    def crosshair_set(self, x=None, y=None):
        """ Sets the crosshair in given coordinates.

        Parameters
        ----------
        x : float, optional
            Crosshair x coordinate.
        y : float, optional
            Crosshair y coordinate.
        """
        self.window = pyds9.DS9(self.window_name)
        if x and y:
            self.window.set('crosshair '+str(x)+' '+str(y)+' physical')
        else:
            self.window.set('mode crosshair')

    def delete_frame(self, all_frames=False):
        """ Deletes last frame or all.

        Parameters
        ----------
        all_frames : bool, optional
            If True all frames are deleted, otherwise only the active one.
        """
        self.window = pyds9.DS9(self.window_name)
        if all_frames:
            self.window.set('frame delete all')
        else:
            self.window.set('frame delete')

    def display(self, *arrays, **kwargs):
        """ Displays one or multiple arrays (listed in ``*arrays``).

        Parameters
        ----------
        *arrays : list of arrays
            The arrays to be displayed.
        **kwargs : dictionary, optional
            Only one parameter is admitted: ``keep_frames``. When
            ``keep_frames`` is passed and is set to True, the new arrays are
            displayed in the existing window (creating new ds9 frames). If no
            ``keep_frames`` is given or if it is set to False, then the existing
            ds9 frames are not preserved.
        """
        self.window = pyds9.DS9(self.window_name)

        if 'keep_frames' in kwargs:
            if not kwargs['keep_frames']:
                self.delete_frame(all_frames=True)
        else:
            self.delete_frame(all_frames=True)

        self.tile()
        for i, array in enumerate(arrays):
            if i == 0:
                self.create_frame()
                self.window.set_np2arr(array)
            else:
                self.window.set('frame new')
                self.window.set_np2arr(array)

    def get(self, paramlist):
        """ XPA get command. Gets data from ds9. See:
        http://ds9.si.edu/doc/ref/xpa.html
        """
        self.window = pyds9.DS9(self.window_name)
        return self.window.get(paramlist)

    def lock(self, scale=True, colorbar=True, crosshair=True, slices=True):
        """ Locks the scaling, colorbar, crosshair position, or slice in cube
        for all existing ds9 frames (wrt the active one). """
        self.window = pyds9.DS9(self.window_name)
        if scale:
            self.window.set('lock scale yes')
        if colorbar:
            self.window.set('lock colorbar yes')
        if crosshair:
            self.window.set('lock crosshair image')
        if slices:
            self.window.set('lock slice image')

    def pan(self, x=None, y=None):
        """ Sets the mode to pan or pans to given coordinates.

        Parameters
        ----------
        x : float, optional
            X coordinate.
        y : float, optional
            Y coordinate.
        """
        self.window = pyds9.DS9(self.window_name)
        if x and y:
            self.window.set('pan to '+str(x)+' '+str(y))
        else:
            self.window.set('mode pan')

    def rotate(self, value=None):
        """ Rotates with a given angle.

        Parameters
        ----------
        value : float, optional
            Angle.
        """
        self.window = pyds9.DS9(self.window_name)
        if value:
            self.window.set('rotate '+str(value))
        else:
            self.window.set('rotate open')

    def scale(self, value=None):
        """ Controls the limits and color scale distribution.

        Parameters
        ----------
        value : str, optional
            Controls the scaling. Syntax:
            scale [linear|log|pow|sqrt|squared|asinh|sinh|histequ]
                  [log exp <value>]
                  [datasec yes|no]
                  [limits <minvalue> <maxvalue>]
                  [mode minmax|<value>|zscale|zmax]
                  [scope local|global]
                  [match]
                  [match limits]
                  [lock [yes|no]]
                  [lock limits [yes|no]]
                  [open|close]
        """
        self.window = pyds9.DS9(self.window_name)
        if value:
            self.window.set('scale '+str(value))
        else:
            self.window.set('scale open')

    def set(self, paramlist, data=None, data_func=None):
        """ XPA set command. Sends data or commands to ds9. See:
        http://ds9.si.edu/doc/ref/xpa.html
        """
        self.window = pyds9.DS9(self.window_name)
        self.window.set(paramlist, data, data_func)

    def tile(self, mode='column'):
        """ Controls the tile display mode.

        Parameters
        ----------
        mode : str, optional
            The mode used for tiling the frames. Syntax:
            tile []
                 [yes|no]
                 [mode grid|column|row]
                 [grid]
                 [grid mode automatic|manual]
                 [grid direction x|y]
                 [grid layout <col> <row>]
                 [grid gap <pixels>]
                 [row]
                 [column]
        """
        self.window = pyds9.DS9(self.window_name)
        if mode == 'off':
            self.window.set('tile off')
        else:
            self.window.set('tile on')
            self.window.set('tile mode ' + mode)

    def unlock(self, scale=True, colorbar=True, crosshair=True, slices=True):
        """ The opposite of the ``lock`` method.
        """
        self.window = pyds9.DS9(self.window_name)
        if scale:
            self.window.set('lock scale no')
        if colorbar:
            self.window.set('lock colorbar no')
        if crosshair:
            self.window.set('lock crosshair none')
        if slices:
            self.window.set('lock slice none')

    def zoom(self, value='to fit'):
        """ Controls the current zoom value for the current frame.

        Parameters
        ----------
        value : str or
            Parameters of the zoom command. Syntax:
            zoom [<value>]
                 [<value> <value>]
                 [to <value>]
                 [to <value> <value>]
                 [in]
                 [out]
                 [to fit]
                 [open|close]
        """
        self.window = pyds9.DS9(self.window_name)
        self.window.set('zoom ' + str(value))



